from typing import Optional, Union

import click
import numpy as np
import pandas as pd
import xarray as xr
from skimage.feature import blob_dog, blob_doh, blob_log

from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Features, Indices, Number, SpotAttributes
from ._base import SpotFinderAlgorithmBase
from .detect import detect_spots, measure_spot_intensity

blob_detectors = {
    'blob_dog': blob_dog,
    'blob_doh': blob_doh,
    'blob_log': blob_log
}


class BlobDetector(SpotFinderAlgorithmBase):

    def __init__(
            self,
            min_sigma: Number,
            max_sigma: Number,
            num_sigma: int,
            threshold: Number,
            overlap: float = 0.5,
            measurement_type='max',
            is_volume: bool = True,
            detector_method: str = 'blob_log'
    ) -> None:
        """Multi-dimensional gaussian spot detector

        This method is a wrapper for skimage.feature.blob_log

        Parameters
        ----------
        min_sigma : float
            The minimum standard deviation for Gaussian Kernel. Keep this low to
            detect smaller blobs.
        max_sigma : float
            The maximum standard deviation for Gaussian Kernel. Keep this high to
            detect larger blobs.
        num_sigma : int
            The number of intermediate values of standard deviations to consider
            between `min_sigma` and `max_sigma`.
        threshold : float
            The absolute lower bound for scale space maxima. Local maxima smaller
            than thresh are ignored. Reduce this to detect blobs with less
            intensities.
        overlap : float [0, 1]
            If two spots have more than this fraction of overlap, the spots are combined
            (default = 0.5)
        measurement_type : str ['max', 'mean']
            name of the function used to calculate the intensity for each identified spot area
        detector_method: str ['blob_dog', 'blob_doh', 'blob_log']
            name of the type of detection method used from skimage.feature

        Notes
        -----
        # TODO ambrosejcarr: revisit after changing dtype assumptions of library to float in [0, 1]
        This spot detector is very sensitive to the threshold that is selected, and the threshold
        is defined as an absolute value -- therefore it must be adjusted depending on the datatype
        of the passed image.

        See Also
        --------
        http://scikit-image.org/docs/dev/auto_examples/features_detection/plot_blob.html

        """
        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.num_sigma = num_sigma
        self.threshold = threshold
        self.overlap = overlap
        self.is_volume = is_volume
        self.measurement_function = self._get_measurement_function(measurement_type)
        try:
            self.detector_method = blob_detectors[detector_method]
        except ValueError:
            "Detector method must be one of {blob_log, blob_dog, blob_doh}"

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """
        Find spots using a gaussian blob finding algorithm

        Parameters
        ----------
        data_image : Union[np.ndarray, xr.DataArray]
            ImageStack containing blobs to be detected

        Returns
        -------
        SpotAttributes :
            DataFrame of metadata containing the coordinates, intensity and radius of each spot

        """

        fitted_blobs_array: np.ndarray = self.detector_method(
            data_image,
            self.min_sigma,
            self.max_sigma,
            self.num_sigma,
            self.threshold,
            self.overlap
        )

        # create the SpotAttributes Table
        columns = [Indices.Z.value, Indices.Y.value, Indices.X.value, Features.SPOT_RADIUS]
        fitted_blobs = pd.DataFrame(data=fitted_blobs_array, columns=columns)

        # convert standard deviation of gaussian kernel used to identify spot to radius of spot
        converted_radius = np.round(fitted_blobs[Features.SPOT_RADIUS] * np.sqrt(3))
        fitted_blobs[Features.SPOT_RADIUS] = converted_radius

        # convert the array to int so it can be used to index
        rounded_blobs = SpotAttributes(fitted_blobs.astype(int))

        rounded_blobs.data['intensity'] = measure_spot_intensity(
            data_image, rounded_blobs, self.measurement_function)
        rounded_blobs.data['spot_id'] = np.arange(rounded_blobs.data.shape[0])

        return rounded_blobs

    def run(
            self,
            data_stack: ImageStack,
            blobs_image: Optional[Union[np.ndarray, xr.DataArray]]=None,
            reference_image_from_max_projection: bool=False,
    ) -> IntensityTable:
        """find spots in an ImageStack

        Parameters
        ----------
        data_stack : ImageStack
            stack containing spots to find
        blobs_image : Union[np.ndarray, xr.DataArray]
        reference_image_from_max_projection : bool
            if True, compute a reference image from the maximum projection of the channels and
            z-planes

        Returns
        -------
        IntensityTable :
            3d tensor containing the intensity of spots across channels and imaging rounds

        """

        intensity_table = detect_spots(
            data_stack=data_stack,
            spot_finding_method=self.image_to_spots,
            reference_image=blobs_image,
            reference_image_from_max_projection=reference_image_from_max_projection,
            measurement_function=self.measurement_function,
            radius_is_gyration=False,
        )

        return intensity_table

    @staticmethod
    @click.command("BlobDetector")
    @click.option(
        "--min-sigma", default=4, type=int, help="Minimum spot size (in standard deviation)")
    @click.option(
        "--max-sigma", default=6, type=int, help="Maximum spot size (in standard deviation)")
    @click.option(
        "--num-sigma", default=20, type=int, help="Number of sigmas to try")
    @click.option(
        "--threshold", default=.01, type=float, help="Dots threshold")
    @click.option(
        "--overlap", default=0.5, type=float,
        help="dots with overlap of greater than this fraction are combined")
    @click.option(
        "--show", default=False, is_flag=True, help="display results visually")
    @click.option(
        "--detector_method", default='blob_log',
        help="str ['blob_dog', 'blob_doh', 'blob_log'] name of the type of "
             "detection method used from skimage.feature"
    )
    @click.pass_context
    def _cli(ctx, min_sigma, max_sigma, num_sigma, threshold, overlap, show, detector_method):
            instance = BlobDetector(min_sigma, max_sigma, num_sigma, threshold, overlap,
                                    detector_method=detector_method)
            #  FIXME: measurement_type, is_volume missing as options; show missing as ctor args
            ctx.obj["component"]._cli_run(ctx, instance)



from typing import Type

import click

from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline import AlgorithmBase, PipelineComponent
from . import fourier_shift
from ._base import RegistrationAlgorithmBase


class Registration(PipelineComponent):

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        return RegistrationAlgorithmBase

    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        stack = ctx.obj["stack"]
        instance.run(stack)
        stack.write(output)

    @staticmethod
    @click.group("registration")
    @click.option("-i", "--input", type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, input, output):
        print("Registering...")
        ctx.obj = dict(
            component=Registration,
            input=input,
            output=output,
            stack=ImageStack.from_path_or_url(input),
        )


Registration._cli_register()



"""
Tests for codebook.code_length method
"""

from starfish import Codebook
from .test_from_code_array import codebook_array_factory


def test_code_length_properly_counts_bit_length_of_codes():
    """
    use the codebook factory from another testing module that produces codes with 3 channels and
    two rounds. This produces a code which should have length 6.

    Then, pass n_round=10 to make the total length 3 * 10 = 30 and verify that the code length is
    returned as 30.
    """
    codebook_data = codebook_array_factory()
    codebook = Codebook.from_code_array(codebook_data)
    assert codebook.code_length == 6

    codebook_long = Codebook.from_code_array(codebook_data, n_round=10)
    assert codebook_long.code_length == 30



import warnings
from typing import Optional

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import pearsonr

from starfish import IntensityTable
from starfish.types import Features
from .util import equalize_axes_extents


def compare_copy_number(
        intensities: IntensityTable,
        other: pd.Series,
        ax: Optional[matplotlib.axes.Axes]=None,
        equalize_axes: bool=False,
        **kwargs,
) -> matplotlib.axes.Axes:
    """
    Plot the copy number of gene expression across all cells against a reference generated by
    another computational approach. Report the $r^2$ value.

    Parameters
    ----------
    intensities : IntensityTable
        decoded IntensityTable
    other : pd.Series
        A pandas series whose index contains gene names and whose values are expression abundances
    ax : Optional[matplotlib.axes.Axes]
        The axis in which to plot the figure. Defaults to the current axis if not provided.
    equalize_axes : bool
        If true, equalize the size of the two axes such that the slope of the comparison can be
        easily interpreted
    kwargs : Dict
        Additional keyword arguments to pass to sns.regplot

    Returns
    -------
    matplotlib.axes.Axes :
        The axis containing the plot.

    """

    ax = ax if ax is not None else plt.gca()
    targets, counts = np.unique(intensities[Features.TARGET].values, return_counts=True)
    this = pd.Series(counts, index=targets)
    tmp = pd.concat([this, other], join='outer', axis=1, sort=True)
    tmp.columns = ['result', 'comparison']

    # this seaborn function does not accept an axis, but we can make sure it draws to the right
    # place by setting the current axis
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FutureWarning)
        sns.regplot('comparison', 'result', data=tmp, ax=ax, **kwargs)

    # calculate the correlation coefficient
    # mask nans
    tmp = tmp.dropna()
    r, _ = pearsonr(tmp['result'], tmp['comparison'])

    plt.text(0.1, 0.85, f'r = {r:.3}', transform=ax.transAxes, fontsize=14)

    if equalize_axes:
        ax = equalize_axes_extents(ax)
    return ax



import unittest

from starfish.util import exec


class TestValidateCommand(unittest.TestCase):

    STAGES = (
        [
            "starfish", "validate", "--help"
        ],
    )

    def test_run_pipeline(self):
        exec.stages(TestValidateCommand.STAGES)



import os
import sys
import unittest

import numpy as np
import pandas as pd

from starfish import IntensityTable
from starfish.test.full_pipelines.cli._base_cli_test import CLITest
from starfish.types import Features


class TestWithDartfishData(CLITest, unittest.TestCase):

    @property
    def subdirs(self):
        return (
            "registered",
            "filtered",
            "results"
        )

    @property
    def stages(self):
        return (
            [
                sys.executable,
                "examples/get_cli_test_data.py",
                "https://d2nhj9g34unfro.cloudfront.net/20181005/DARTFISH-TEST/",
                lambda tempdir, *args, **kwargs: os.path.join(tempdir, "registered")
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered/fov_001", "hybridization.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "scale_filtered.json"),
                "ScaleByPercentile",
                "--p", "100",
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "scale_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "zero_filtered.json"),
                "ZeroByChannelMagnitude",
                "--thresh", ".05",
            ],
            [
                "starfish", "detect_spots",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "zero_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "spots.nc"),
                "--codebook", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered", "codebook.json"),
                "PixelSpotDetector",
                "--distance-threshold", "3",
                "--magnitude-threshold", ".5",
                "--min-area", "5",
                "--max-area", "30",
            ],
        )

    def verify_results(self, intensities):
        assert intensities[Features.PASSES_THRESHOLDS].sum()

        spots_df = IntensityTable(
            intensities.where(intensities[Features.PASSES_THRESHOLDS], drop=True)
        ).to_features_dataframe()
        spots_df['area'] = np.pi * spots_df['radius'] ** 2

        # verify number of spots detected
        spots_passing_filters = intensities[Features.PASSES_THRESHOLDS].sum()
        assert spots_passing_filters == 53  # TODO note, had to change this by 1

        # compare to benchmark data -- note that this particular part of the dataset
        # appears completely uncorrelated
        cnts_benchmark = pd.read_csv(
            'https://d2nhj9g34unfro.cloudfront.net/20181005/DARTFISH/fov_001/counts.csv')

        min_dist = 0.6
        cnts_starfish = spots_df[spots_df.distance <= min_dist].groupby('target').count()['area']
        cnts_starfish = cnts_starfish.reset_index(level=0)
        cnts_starfish.rename(columns={'target': 'gene', 'area': 'cnt_starfish'}, inplace=True)

        # get top 5 genes and verify they are correct
        high_expression_genes = cnts_starfish.sort_values('cnt_starfish', ascending=False).head(5)

        assert np.array_equal(
            high_expression_genes['cnt_starfish'].values,
            [7, 3, 2, 2, 2]
        )
        assert np.array_equal(
            high_expression_genes['gene'].values,
            ['MBP', 'MOBP', 'ADCY8', 'TRIM66', 'SYT6']
        )

        # verify correlation is accurate for this subset of the image
        benchmark_comparison = pd.merge(cnts_benchmark, cnts_starfish, on='gene', how='left')
        benchmark_comparison.head(20)

        x = benchmark_comparison.dropna().cnt.values
        y = benchmark_comparison.dropna().cnt_starfish.values
        corrcoef = np.corrcoef(x, y)
        corrcoef = corrcoef[0, 1]

        assert np.round(corrcoef, 5) == 0.04422



"""
These tests ensure that we can share an np-array or something that encapsulates an np-array among
processes.  The data is shared such that any process can write to the array and it is visible to all
the other processes.  It is up to the developer that uses this mechanism to ensure data integrity is
maintained.
"""
import copy
import ctypes
import multiprocessing
import warnings
from functools import partial
# Even though we import multiprocessing, mypy can't find the Array class.  To avoid sprinkling
# ignore markers all over the file, we explicitly import the symbol and put the ignore marker here.
from multiprocessing import Array as mp_array  # type: ignore
from typing import Any, Callable, Tuple

import numpy as np
import xarray as xr

from starfish import ImageStack
from starfish.imagestack import _mp_dataarray
from starfish.multiprocessing.shmem import SharedMemory


def test_numpy_array(nitems: int=10):
    """
    Try to share a numpy array directly.  This should fail as numpy is by default a copy-on-write
    object.  The worker process will write to array and we should not see the changes.
    """
    array = np.zeros(shape=(nitems,), dtype=np.uint8)
    _start_process_to_test_shmem(
        array_holder=array,
        decoder=_decode_numpy_array_to_numpy_array,
        nitems=nitems)
    for ix in range(nitems):
        assert array[ix] == 0


def _decode_numpy_array_to_numpy_array(array):
    """
    When passing naked numpy arrays into subprocesses, no work is needed to retrieve the array. See
    :py:method:`_decode_array_to_numpy_array` for an example where decoding is necessary.
    """
    return array


def test_shmem_numpy_array(nitems: int=10):
    """
    Try to share a numpy array based on a multiprocessing Array object.  The array object is passed
    to the worker process.  The worker process reconstitutes the numpy array from that memory buffer
    and writes to the reconstituted numpy array.  Writes in the worker process should be visible in
    the parent process.
    """
    buffer = mp_array(ctypes.c_uint8, nitems)
    array = _decode_array_to_numpy_array(buffer)
    array.fill(0)
    _start_process_to_test_shmem(
        array_holder=buffer,
        decoder=_decode_array_to_numpy_array,
        nitems=nitems)
    for ix in range(nitems):
        assert array[ix] == ix


def _decode_array_to_numpy_array(array):
    return np.frombuffer(array.get_obj(), dtype=np.uint8)


class TestWrappedArray:
    """
    Dummy class that wraps a multiprocessing array in an object.
    """
    def __init__(self, array: mp_array) -> None:
        self.array = array


def test_wrapped_shmem_numpy_array(nitems: int=10):
    """
    Try to share a numpy array based on a multiprocessing Array object.  The array object is wrapped
    in a container class and passed to the worker process.  The worker process reconstitutes the
    numpy array from that memory buffer and writes to the reconstituted numpy array.  Writes in the
    worker process should be visible in the parent process.
    """
    buffer = mp_array(ctypes.c_uint8, nitems)
    array = _decode_array_to_numpy_array(buffer)
    array.fill(0)
    _start_process_to_test_shmem(
        array_holder=TestWrappedArray(buffer),
        decoder=_decode_wrapped_array_to_numpy_array,
        nitems=nitems)
    for ix in range(nitems):
        assert array[ix] == ix


def _decode_wrapped_array_to_numpy_array(wrapped_array: TestWrappedArray) -> np.ndarray:
    return np.frombuffer(wrapped_array.array.get_obj(), dtype=np.uint8)


def test_xr_deepcopy(nitems: int=10) -> None:
    """
    Instantiate an :py:class:`xarray.DataArray` and run
    :py:method:`starfish.imagestack.mp_dataarray.xr_deepcopy` on it.  The copy is passed to worker
    processes.  Worker processes reconstitute a numpy array from the buffer and attempts to writes
    to the numpy array.  Writes in the worker process should be visible in the parent process in the
    copy but not the original.
    """
    shape = (nitems, )
    source = np.zeros(shape, dtype=np.uint8)
    dataarray = xr.DataArray(source)
    copy, buffer = _mp_dataarray.xr_deepcopy(dataarray)
    _start_process_to_test_shmem(
        array_holder=buffer,
        decoder=_decode_array_to_numpy_array,
        nitems=nitems)
    for ix in range(nitems):
        assert dataarray[ix] == 0
        assert copy[ix] == ix


def test_imagestack_deepcopy(nitems: int=10) -> None:
    """
    Instantiate an :py:class:`ImageStack` and deepcopy it.  Worker processes reconstitute a numpy
    array from the buffer and attempts to writes to the numpy array.  Writes in the worker process
    should be visible in the parent process.
    """
    shape = (nitems, 3, 4, 5, 6)
    dtype = np.float32
    source = np.zeros(shape, dtype=np.float32)
    imagestack = ImageStack.from_numpy_array(source)
    imagestack_copy = copy.deepcopy(imagestack)
    _start_process_to_test_shmem(
        array_holder=imagestack_copy._data._backing_mp_array,
        decoder=partial(_decode_imagestack_array_to_numpy_array, shape, dtype),
        nitems=nitems)
    for ix in range(nitems):
        assert (imagestack.xarray[ix] == 0).all()
        assert np.allclose(imagestack_copy.xarray[ix], ix)


def _decode_imagestack_array_to_numpy_array(
        shape: Tuple[int, ...], dtype, buffer) -> np.ndarray:
    unshaped_np_array = np.frombuffer(buffer.get_obj(), dtype=dtype)
    return unshaped_np_array.reshape(shape)


def test_imagestack_xarray_deepcopy(nitems: int=10) -> None:
    """
    Instantiate an :py:class:`ImageStack` and deepcopy the xarray directly.  This should work, but
    prompt a warning.
    """
    shape = (nitems, 3, 4, 5, 6)
    dtype = np.float32
    source = np.zeros(shape, dtype=dtype)
    imagestack = ImageStack.from_numpy_array(source)
    with warnings.catch_warnings(record=True) as warnings_:
        copy.deepcopy(imagestack.xarray)
        assert len(warnings_) == 1  # type: ignore


def _write_to_shared_memory_array(decoder: Callable[[Any], np.ndarray], position: int) -> None:
    array = decoder(SharedMemory.get_payload())
    array[position] = position


def _start_process_to_test_shmem(
        *,
        array_holder: Any,
        decoder: Callable[[Any], np.ndarray],
        nitems: int) -> None:
    """
    This method leverages multiprocessing.Pool to test process initializers and shared memory array
    decoders.

    Starts a process Pool and passes :py:method:`SharedMemory.initializer` as its initializer, which
    will store `array_holder` as a global variable inside the initialized process.

    Creates `bound_func`, which will be executed by the subprocesses of Pool.  `bound_func` is a
    partial function constructed from :py:method:`_write_to_shared_memory_array` that exposes
    `decoder`, a method that gets a numpy array out of array_holder and then writes a value to it
    (inside the spawned process).

    This enables writing of tests that check whether the written values are available in the parent
    processes, which depends on `SharedMemory.initializer`, `array_holder`, and `decoder`.

    All the subprocesses must complete before this method returns.

    Parameters
    ----------
    array_holder : Any
        An object that when passed to the callable specified in `decoder`.  This is the object that
        is passed in :py:class:`multiprocessing.Pool`'s initargs.
    decoder : Callable[[Any], np.ndarray]
        A callable that takes the object in `array_holder` and returns a np.ndarray.
    nitems : int
        Number of jobs to spawn.  Each job will write a single element of the array.
    """
    bound_func = partial(_write_to_shared_memory_array, decoder)

    vals = [ix for ix in range(nitems)]
    with multiprocessing.Pool(
            initializer=SharedMemory.initializer, initargs=(array_holder,)) as pool:
        pool.map(bound_func, vals)



from collections import OrderedDict
from typing import Union

import numpy as np

from starfish import IntensityTable
from starfish.imagestack import physical_coordinate_calculator
from starfish.intensity_table import intensity_table_coordinates
from starfish.test import test_utils
from starfish.types import Coordinates, Indices, PhysicalCoordinateTypes


NUMBER_SPOTS = 10


def physical_cord_to_pixel_value(physical_coord: float,
                                 physical_pixel_size: Union[float, int],
                                 coordinates_at_pixel_offset_0: int):

    return (physical_coord - coordinates_at_pixel_offset_0) / physical_pixel_size


def test_tranfering_physical_coords_to_intensity_table():
    stack_shape = OrderedDict([(Indices.ROUND, 3), (Indices.CH, 2),
                               (Indices.Z, 1), (Indices.Y, 50), (Indices.X, 40)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, 1),
                                   (PhysicalCoordinateTypes.X_MAX, 2),
                                   (PhysicalCoordinateTypes.Y_MIN, 4),
                                   (PhysicalCoordinateTypes.Y_MAX, 6),
                                   (PhysicalCoordinateTypes.Z_MIN, 1),
                                   (PhysicalCoordinateTypes.Z_MAX, 3)])

    stack = test_utils.imagestack_with_coords_factory(stack_shape, physical_coords)
    codebook = test_utils.codebook_array_factory()

    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=stack_shape[Indices.Z],
        height=stack_shape[Indices.Y],
        width=stack_shape[Indices.X],
        n_spots=NUMBER_SPOTS
    )

    intensities = intensity_table_coordinates.\
        transfer_physical_coords_from_imagestack_to_intensity_table(stack, intensities)

    # Assert that new cords were added
    xc = intensities.coords[Coordinates.X]
    yc = intensities.coords[Coordinates.Y]
    zc = intensities.coords[Coordinates.Z]
    assert xc.size == NUMBER_SPOTS
    assert yc.size == NUMBER_SPOTS
    assert zc.size == NUMBER_SPOTS

    physical_pixel_size_x = physical_coordinate_calculator._calculate_physical_pixel_size(
        coord_min=physical_coords[PhysicalCoordinateTypes.X_MIN],
        coord_max=physical_coords[PhysicalCoordinateTypes.X_MAX],
        num_pixels=stack_shape[Indices.X])

    physical_pixel_size_y = physical_coordinate_calculator._calculate_physical_pixel_size(
        coord_min=physical_coords[PhysicalCoordinateTypes.Y_MIN],
        coord_max=physical_coords[PhysicalCoordinateTypes.Y_MAX],
        num_pixels=stack_shape[Indices.Y])

    # Assert that the physical coords align with their corresponding pixel coords
    for spot in xc.features:
        pixel_x = spot[Indices.X.value].data
        physical_x = spot[Coordinates.X.value].data
        calculated_pixel = physical_cord_to_pixel_value(physical_x,
                                                        physical_pixel_size_x,
                                                        physical_coords[
                                                            PhysicalCoordinateTypes.X_MIN
                                                        ])
        assert np.isclose(pixel_x, calculated_pixel)

    for spot in yc.features:
        pixel_y = spot[Indices.Y.value].data
        physical_y = spot[Coordinates.Y.value].data
        calculated_pixel = physical_cord_to_pixel_value(physical_y,
                                                        physical_pixel_size_y,
                                                        physical_coords[
                                                            PhysicalCoordinateTypes.Y_MIN
                                                        ])
        assert np.isclose(pixel_y, calculated_pixel)

    # Assert that zc value is middle of z range
    for spot in zc.features:
        physical_z = spot[Coordinates.Z.value].data
        assert np.isclose(physical_coords[PhysicalCoordinateTypes.Z_MAX],
                          (physical_z * 2) - physical_coords[PhysicalCoordinateTypes.Z_MIN])



# deal with numpy import warnings due to cython
# See: https://stackoverflow.com/questions/40845304/
#      runtimewarning-numpy-dtype-size-changed-may-indicate-binary-incompatibility)
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")  # noqa
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")  # noqa

# image processing methods and objects
from . import image
# spot detection and manipulation
from . import spots
# top-level objects
from .codebook.codebook import Codebook
from .experiment.experiment import Experiment, FieldOfView
from .imagestack.imagestack import ImageStack
from .intensity_table.intensity_table import IntensityTable
from .starfish import starfish

if __name__ == "__main__":
    starfish()



"""
Tests for IntensityTable.from_image_stack method
"""

import numpy as np

from starfish import ImageStack, IntensityTable
from starfish.types import Indices


# TODO ambrosejcarr: crop is not tested because it should be moved out of this function
def test_intensity_table_can_be_constructed_from_an_imagestack():
    """
    ImageStack has enough information to create an IntensityTable without additional SpotAttributes.
    Each feature is a pixel, and therefore the SpotAttributes can be extracted from the relative
    locations.
    """
    r, c, z, y, x = 1, 5, 2, 2, 5
    data = np.zeros(100, dtype=np.float32).reshape(r, c, z, y, x)
    image_stack = ImageStack.from_numpy_array(data)
    intensities = IntensityTable.from_image_stack(image_stack)

    # there should be 100 features
    assert np.product(intensities.shape) == 100

    # the max features should be equal to the array extent (2, 2, 5) minus one, since indices
    # are being compared and python is zero based
    # import pdb; pdb.set_trace()
    assert np.max(intensities[Indices.Z.value].values) == z - 1
    assert np.max(intensities[Indices.Y.value].values) == y - 1
    assert np.max(intensities[Indices.X.value].values) == x - 1

    # the number of channels and hybridizations should match the ImageStack
    assert intensities.sizes[Indices.CH.value] == c
    assert intensities.sizes[Indices.ROUND.value] == r



import collections
from typing import Mapping, Optional, Type

from .algorithmbase import AlgorithmBase


class PipelineComponentType(type):
    """
    This is the metaclass for PipelineComponent.  As each subclass that is _not_ PipelineComponent
    is created, it sets up a map between the algorithm name and the class that implements it.
    """
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if len(bases) != 0:
            # this is _not_ PipelineComponent.  Instead, it's a subclass of PipelineComponent.
            PipelineComponentType._ensure_algorithms_setup(cls)

    @classmethod
    def _ensure_algorithms_setup(mcs, cls):
        if cls._algorithm_to_class_map_int is None:
            cls._algorithm_to_class_map_int = dict()

            queue = collections.deque(cls._get_algorithm_base_class().__subclasses__())
            while len(queue) > 0:
                algorithm_cls = queue.popleft()
                queue.extend(algorithm_cls.__subclasses__())

                cls._algorithm_to_class_map_int[algorithm_cls.__name__] = algorithm_cls

                setattr(cls, algorithm_cls._get_algorithm_name(), algorithm_cls)


class PipelineComponent(metaclass=PipelineComponentType):

    _algorithm_to_class_map_int: Optional[Mapping[str, Type]] = None

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        """
        Get the base class that algorithms which implement this pipeline stage must extend.
        Pipeline components must provide this method.
        """
        raise NotImplementedError()

    @classmethod
    def _algorithm_to_class_map(cls):
        """Returns a mapping from algorithm names to the classes that implement them."""
        return cls._algorithm_to_class_map_int

    @classmethod
    def _cli_run(cls, ctx, instance, *args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _cli_register(cls):
        for algorithm_cls in cls._algorithm_to_class_map().values():
            cls._cli.add_command(algorithm_cls._cli)



"""
Tests for codebook.per_round_max_decode method
"""

import numpy as np
import pandas as pd
import pytest

from starfish import Codebook, IntensityTable
from starfish.types import Features, Indices, SpotAttributes


def intensity_table_factory(data: np.ndarray=np.array([[[0, 3], [4, 0]]])) -> IntensityTable:
    """IntensityTable with a single feature that was measured over 2 channels and 2 rounds."""

    # generates spot attributes equal in size to the number of passed features.
    # each attribute has coordinates (z, y, x) equal to the feature index, and radius 1.
    spot_attributes_data = pd.DataFrame(
        data=np.array([[i, i, i, 1] for i in np.arange(data.shape[0])]),
        columns=[Indices.Z, Indices.Y, Indices.X, Features.SPOT_RADIUS]
    )

    spot_attributes = SpotAttributes(spot_attributes_data)
    intensity_table = IntensityTable.from_spot_data(data, spot_attributes)
    return intensity_table


def codebook_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    codebook_array = [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 1, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_A'
        },
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 1, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_B'
        },
    ]
    return Codebook.from_code_array(codebook_array)


def test_intensity_tables_with_different_nubmers_of_codes_or_channels_throw_value_error():
    """
    The test passes a 3-round and 1-round IntensityTable to a 2-round codebook. Both should
    raise a ValueError.
    """
    data = np.array(
        [[[4, 3, 1],
          [4, 0, 2]]]
    )
    codebook = codebook_factory()
    intensities = intensity_table_factory(data)
    with pytest.raises(ValueError):
        codebook.decode_per_round_max(intensities)

    with pytest.raises(ValueError):
        codebook.decode_per_round_max(intensities.sel(round=1))


def test_example_intensity_decodes_to_gene_a():
    """The single intensity should decode to gene a. Verify that it does."""
    intensities = intensity_table_factory()
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)

    assert decoded_intensities[Features.TARGET] == 'GENE_A'
    assert np.array_equal(decoded_intensities[Features.PASSES_THRESHOLDS], [True])
    assert np.array_equal(decoded_intensities[Features.DISTANCE], [0])


def test_example_intensity_that_has_no_match_is_assigned_nan():
    """
    The decoder assigns nan when it doesn't receive a match, but the nan is converted to a
    string due to the object dtype
    """
    no_match = np.array([[[3, 0], [0, 4]]])
    intensities = intensity_table_factory(no_match)
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)

    # even though the feature did not decode, it should still be returned
    assert decoded_intensities.sizes[Features.AXIS] == 1

    # assert that the feature did not decode
    assert decoded_intensities[Features.TARGET] == 'nan'
    assert np.array_equal(decoded_intensities[Features.PASSES_THRESHOLDS], [False])

    # distance is calculated as the fraction of signal NOT in the selected channel. Here all of
    # the signal is in the max channel, so distance is 0.
    # Admittedly, this is a little bit confusing, since distance here measures both how correct
    # a code is, but also how wrong one is when it _doesn't_ decode properly.
    assert np.array_equal(decoded_intensities[Features.DISTANCE], [0])


def test_argmax_selects_the_last_equal_intensity_channel_and_decodes_consistently():
    """
    When two channels are tied, argmax selects the last first one. This is the code for Gene A:

    [[0, 1],
     [1, 0]]

    both of these observations could potentially match the code for GENE_A:

    [[4, 3],
     [4, 0]]

    and

    [[0, 3],
     [4, 3]]

    The first is interpreted as argmax as:

    [[1, 1],
     [0, 0]]

    which fails to match, while the second is interpreted as

    [[0, 1],
     [1, 0]]

    and matches GENE_A.

    This test also verifies that the decoder can operate on multiple codes.
    """

    data = np.array(
        [[[4, 3],  # this code is decoded "wrong"
          [4, 0]],
         [[0, 3],  # this code is decoded "right"
          [4, 3]]]
    )
    intensities = intensity_table_factory(data)
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)
    assert np.array_equal(decoded_intensities[Features.TARGET].values, ['nan', 'GENE_A'])



import json
import posixpath
import sys
from typing import Dict

import click
from pkg_resources import resource_filename
from slicedimage.io import resolve_path_or_url

from .util import SpaceTxValidator


def _get_absolute_schema_path(schema_name: str) -> str:
    """turn the name of the schema into an absolute path by joining it to <package_root>/schema."""
    return resource_filename("sptx_format", posixpath.join("schema", schema_name))


@click.command()
@click.option('--experiment-json', help='image metadata file to validate')
def validate_sptx(experiment_json: str, fuzz: bool=False) -> bool:
    return validate(experiment_json, fuzz)

def validate(experiment_json: str, fuzz: bool=False) -> bool:
    """validate a spaceTx formatted experiment.
    Accepts local filepaths or files hosted at http links.

    Parameters
    ----------
    experiment_json : str
        path or URL to a target json object to be validated against the schema passed to this
        object's constructor
    fuzz : bool
        whether or not to perform element-by-element fuzzing.
        If true, will return true and will *not* use warnings.

    Returns
    -------
    bool :
        True, if object valid or fuzz=True, else False
    """

    valid = True

    # use slicedimage to read the top-level experiment json file passed by the user
    backend, name, baseurl = resolve_path_or_url(experiment_json)
    with backend.read_contextmanager(name) as fh:
        experiment = json.load(fh)

    # validate experiment.json
    experiment_validator = SpaceTxValidator(_get_absolute_schema_path('experiment.json'))
    if fuzz:
        experiment_validator.fuzz_object(experiment, name)
    else:
        valid &= experiment_validator.validate_object(experiment, name)

    # loop over all the manifests that are stored in images. Disallowed names will have already been
    # excluded by experiment validation.
    manifests = []
    for manifest in experiment['images'].values():
        with backend.read_contextmanager(manifest) as fh:
            manifests.append((json.load(fh), manifest))

    fovs = []
    manifest_validator = SpaceTxValidator(_get_absolute_schema_path('fov_manifest.json'))
    for manifest, filename in manifests:
        if fuzz:
            manifest_validator.fuzz_object(manifest, filename)
        else:
            if not manifest_validator.validate_object(manifest, filename):
                valid = False
            else:
                for key, fov in manifest['contents'].items():
                    with backend.read_contextmanager(fov) as fh:
                        fovs.append((json.load(fh), fov))

    # fovs may be empty if the manifests were not valid
    if fovs:
        fov_schema = _get_absolute_schema_path('field_of_view/field_of_view.json')
        fov_validator = SpaceTxValidator(fov_schema)
        for fov, filename in fovs:
            if fuzz:
                fov_validator.fuzz_object(fov, filename)
            else:
                valid &= fov_validator.validate_object(fov, filename)

    # validate codebook
    codebook_validator = SpaceTxValidator(_get_absolute_schema_path('codebook/codebook.json'))
    codebook_file = experiment.get('codebook')
    codebook: Dict = {}
    if codebook_file is not None:
        with backend.read_contextmanager(codebook_file) as fh:
            codebook = json.load(fh)
    if fuzz:
        codebook_validator.fuzz_object(codebook, codebook_file)
    else:
        valid &= codebook_validator.validate_object(codebook, codebook_file)

    return valid


if __name__ == "__main__":
    valid = validate_sptx()

    if valid:
        sys.exit(0)
    else:
        sys.exit(1)



"""
Tests for CombineAdjacentFeatures._create_spot_attributes method
"""

import numpy as np
from skimage.measure import regionprops

from starfish.spots._detector.combine_adjacent_features import CombineAdjacentFeatures, TargetsMap
from starfish.types import Features, Indices, SpotAttributes
from .test_calculate_mean_pixel_traces import labeled_intensities_factory  # reuse this fixture


def test_create_spot_attributes():
    """
    Verify that SpotAttributes are correctly created by verifying that a simple example maps each
    feature to the expected x, y, z coordinates and that the correct targets are associated with
    each feature
    """
    # make some fixtures
    intensity_table, label_image, decoded_image = labeled_intensities_factory()
    region_properties = regionprops(label_image)
    target_map = TargetsMap(np.array(list('abcdef')))
    caf = CombineAdjacentFeatures(min_area=1, max_area=3, connectivity=2)
    spot_attributes, passes_filters = caf._create_spot_attributes(
        region_properties, decoded_image, target_map
    )

    assert isinstance(spot_attributes, SpotAttributes)

    # should have a shape equal to the number of features (4)
    assert spot_attributes.data.shape[0] == 4

    # each feature should be correctly localized to z, y, x, and the SpotAttributes should be
    # sorted in that order
    assert np.array_equal(spot_attributes.data[Indices.Z], [0, 0, 0, 0])
    assert np.array_equal(spot_attributes.data[Indices.Y], [0, 0, 1, 1])
    assert np.array_equal(spot_attributes.data[Indices.X], [0, 1, 0, 1])

    # spots should map to edcb -- the decoded image has 4 values: [5, 4, 3, 2]. The target map
    # starts at np.nan, then counts sequentially from 1. Thus, 2 should map to b, and from there
    # the values are sequential. f=6 is not present.
    assert np.array_equal(spot_attributes.data[Features.TARGET].values, list('edcb'))



import pytest

from starfish.experiment.builder.defaultproviders import OnesTile, tile_fetcher_factory
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices

NUM_ROUND = 4
NUM_CH = 2
NUM_Z = 12


class OnesTilesWithExtras(OnesTile):
    def __init__(self, extras: dict, *args, **kwargs) -> None:
        super().__init__((10, 10))
        self._extras = extras

    @property
    def extras(self):
        return self._extras


def test_metadata():
    """
    Normal situation where all the tiles have uniform keys for both indices and extras.
    """
    tile_fetcher = tile_fetcher_factory(
        OnesTilesWithExtras,
        False,
        {
            'random_key': {
                'hello': "world",
            }
        }
    )

    stack = ImageStack.synthetic_stack(
        num_round=NUM_ROUND, num_ch=NUM_CH, num_z=NUM_Z, tile_fetcher=tile_fetcher,
    )
    table = stack.tile_metadata
    assert len(table) == NUM_ROUND * NUM_CH * NUM_Z


def test_missing_extras():
    """
    If the extras are not present on some of the tiles, it should still work.
    """
    class OnesTilesWithExtrasMostly(OnesTile):
        def __init__(self, fov, r, ch, z, extras: dict) -> None:
            super().__init__((10, 10))
            self.fov = fov
            self._extras = extras

        @property
        def extras(self):
            if self.fov == 0:
                return None
            return self._extras

    tile_fetcher = tile_fetcher_factory(
        OnesTilesWithExtrasMostly,
        True,
        {
            'random_key': {
                'hello': "world",
            }
        }
    )

    stack = ImageStack.synthetic_stack(
        num_round=NUM_ROUND, num_ch=NUM_CH, num_z=NUM_Z, tile_fetcher=tile_fetcher,
    )
    table = stack.tile_metadata
    assert len(table) == NUM_ROUND * NUM_CH * NUM_Z


def test_conflict():
    """
    Tiles that have extras that conflict with indices should produce an error.
    """
    tile_fetcher = tile_fetcher_factory(
        OnesTilesWithExtras,
        False,
        {
            Indices.ROUND: {
                'hello': "world",
            }
        }
    )

    stack = ImageStack.synthetic_stack(
        num_round=NUM_ROUND, num_ch=NUM_CH, num_z=NUM_Z, tile_fetcher=tile_fetcher,
    )
    with pytest.raises(ValueError):
        stack.tile_metadata



import json

import click

from starfish.types import Indices
from . import AUX_IMAGE_NAMES, write_experiment_json


class StarfishIndex(click.ParamType):

    name = "starfish-index"

    def convert(self, spec_json, param, ctx):
        try:
            spec = json.loads(spec_json)
        except json.decoder.JSONDecodeError:
            self.fail(
                "Could not parse {} into a valid index specification.".format(spec_json))

        return {
            Indices.ROUND: spec.get(Indices.ROUND, 1),
            Indices.CH: spec.get(Indices.CH, 1),
            Indices.Z: spec.get(Indices.Z, 1),
        }

def dimensions_option(name, required):
    return click.option(
        "--{}-dimensions".format(name),
        type=StarfishIndex(), required=required,
        help="Dimensions for the {} images.  Should be a json dict, with {}, {}, "
             "and {} as the possible keys.  The value should be the shape along that "
             "dimension.  If a key is not present, the value is assumed to be 0."
             .format(
             name,
             Indices.ROUND.value,
             Indices.CH.value,
             Indices.Z.value))

decorators = [
    click.command(),
    click.argument("output_dir", type=click.Path(exists=True, file_okay=False, writable=True)),
    click.option("--fov-count", type=int, required=True, help="Number of FOVs in this experiment."),
    dimensions_option("hybridization", True),
]
for image_name in AUX_IMAGE_NAMES:
    decorators.append(dimensions_option(image_name, False))

def build(output_dir, fov_count, hybridization_dimensions, **kwargs):
    write_experiment_json(
        output_dir, fov_count, hybridization_dimensions,
        kwargs
    )

for decorator in reversed(decorators):
    build = decorator(build)



"""
Because multiprocessing.Pool demands that any shared-memory constructs be passed in as initializer
arguments, and not as arguments to multiprocessing.Pool.map or multiprocessing.Pool.apply, we need a
stub initializer that accepts the shared-memory construct and store it in a global for retrieval.

Does not work:
  arrays = [multiprocessing.Array(..)]
  with multiprocessing.Pool() as pool:
    pool.map(some_func, arrays)

Does work:
  array = multiprocessing.Array(..)
  data = range(10)
  with multiprocessing.Pool(initializer=some_function, initargs=(array,)) as pool:
    pool.map(some_func, data)

This wraps the ugliness of a global variable inside a class that's only used for this purpose.  An
example of how this might be used would be:

  def worker(val):
    array = SharedMemory.get_payload()
    # do something with the shared memory

  array = multiprocessing.Array(..)
  data = range(10)
  with multiprocessing.Pool(initializer=SharedMemory.initalizer, initargs=(array,)) as pool:
    pool.map(worker, data)
"""
from typing import Any


class SharedMemory:
    _payload = None

    @staticmethod
    def initializer(payload: Any) -> None:
        SharedMemory._payload = payload

    @staticmethod
    def get_payload() -> Any:
        return SharedMemory._payload



from collections import OrderedDict

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices


def test_imagestack_indexing():
    """Tests indexing on an Imagestack with a shape (5, 5, 15, 200, 200)
        steps:
        1.) stack.sel(indexers)
        2.) assert new shape of stack is what we expect

    """

    stack = ImageStack.synthetic_stack(num_round=5, num_ch=5, num_z=15,
                                       tile_height=200, tile_width=200)

    # index on range of rounds and single ch and Z
    indexed = stack.sel({Indices.ROUND: (1, None), Indices.CH: 0, Indices.Z: 0})
    expected_shape = OrderedDict([(Indices.ROUND, 4), (Indices.CH, 1),
                                  (Indices.Z, 1), (Indices.Y, 200), (Indices.X, 200)])
    assert indexed.shape == expected_shape

    # index on single round ch and z
    indexed = stack.sel({Indices.ROUND: 0, Indices.CH: 0, Indices.Z: 0})
    expected_shape = OrderedDict([(Indices.ROUND, 1), (Indices.CH, 1),
                                  (Indices.Z, 1), (Indices.Y, 200), (Indices.X, 200)])
    assert indexed.shape == expected_shape

    # index on single round and range of ch
    indexed = stack.sel({Indices.ROUND: 1, Indices.CH: (3, None)})
    expected_shape = OrderedDict(
        [(Indices.ROUND, 1), (Indices.CH, 2), (Indices.Z, 15), (Indices.Y, 200), (Indices.X, 200)])
    assert indexed.shape == expected_shape

    # index on single round and range of ch and Z
    indexed = stack.sel({Indices.ROUND: 1, Indices.CH: (None, 3), Indices.Z: (7, None)})
    expected_shape = OrderedDict(
        [(Indices.ROUND, 1), (Indices.CH, 3), (Indices.Z, 8), (Indices.Y, 200), (Indices.X, 200)])
    assert indexed.shape == expected_shape

    # index on first half of X and single value of Y
    indexed_stack = stack.sel({Indices.ROUND: 0, Indices.CH: 0, Indices.Z: 1,
                               Indices.Y: 100, Indices.X: (None, 100)})
    expected_shape = OrderedDict([(Indices.ROUND, 1), (Indices.CH, 1),
                                  (Indices.Z, 1), (Indices.Y, 1), (Indices.X, 100)])
    assert indexed_stack.shape == expected_shape

    # index on first half of X and Y
    indexed_stack = stack.sel({Indices.Y: (None, 100), Indices.X: (None, 100)})

    expected_shape = OrderedDict([(Indices.ROUND, 5), (Indices.CH, 5),
                                  (Indices.Z, 15), (Indices.Y, 100), (Indices.X, 100)])
    assert indexed_stack.shape == expected_shape

    # index on single x and y
    indexed_stack = stack.sel({Indices.ROUND: 0, Indices.CH: 0, Indices.Z: 1,
                               Indices.Y: 100, Indices.X: 150})
    expected_shape = OrderedDict([(Indices.ROUND, 1), (Indices.CH, 1),
                                  (Indices.Z, 1), (Indices.Y, 1), (Indices.X, 1)])
    assert indexed_stack.shape == expected_shape

    # Negative indexing
    indexed_stack = stack.sel({Indices.ROUND: 0, Indices.CH: 0, Indices.Z: 1,
                               Indices.Y: (None, -10), Indices.X: (None, -10)})
    expected_shape = OrderedDict([(Indices.ROUND, 1), (Indices.CH, 1),
                                  (Indices.Z, 1), (Indices.Y, 190), (Indices.X, 190)])
    assert indexed_stack.shape == expected_shape



from typing import Any, Mapping, Optional, Tuple, Union

import xarray as xr

from starfish.imagestack import indexing_utils
from starfish.types import (
    Coordinates,
    Indices,
    Number,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes
)


def calc_new_physical_coords_array(
        physical_coordinates: xr.DataArray,
        stack_shape: Mapping[Indices, int],
        indexers: Mapping[str, Union[int, slice]],
) -> xr.DataArray:
    """Calculates the resulting coordinates array from indexing on each dimension in indexers

    Parameters
    ----------
    physical_coordinates: xarray
        xarray that holds the min/max values of physical coordinates per tile (Round, Ch, Z).
    stack_shape : Dict[str, int]
        The shape of the image tensor by categorical index (channels, imaging rounds, z-layers)
    indexers : Dict[str, (int/slice)]
        A dictionary of dim:index where index is the value or range to index the dimension

    Returns
    -------
    A coordinates xarray indexed by R, CH, V and values recalculated according to indexing on X/Y
    """

    new_coords = physical_coordinates.copy(deep=True)
    # index by R, CH, V
    key = {Indices.ROUND.value: indexers[Indices.ROUND.value],
           Indices.CH.value: indexers[Indices.CH.value],
           Indices.Z.value: indexers[Indices.Z.value]}
    new_coords = indexing_utils.index_keep_dimensions(new_coords, key)
    # check if X or Y dimension indexed, if so recalculate physcal coordinate min/max values
    if _needs_coords_recalculating(indexers[Indices.X.value], indexers[Indices.Y.value]):
        _recalculate_physical_coordinate_ranges(stack_shape, indexers, new_coords)
    return new_coords


def _needs_coords_recalculating(
        x_indexers: Union[int, slice], y_indexers: Union[int, slice]) -> bool:
    if isinstance(x_indexers, int) or isinstance(y_indexers, int):
        return True
    return not (x_indexers.start is x_indexers.stop is y_indexers.start is y_indexers.stop is None)


def _recalculate_physical_coordinate_ranges(
        stack_shape: Mapping[Indices, int],
        indexers: Mapping[str, Union[int, slice]],
        coords_array: xr.DataArray,
) -> None:
    """Iterates through coordinates array and recalculates x, y min/max physical coordinate values
     based on how the x and y dimensions were indexed

    Parameters
    ----------
    indexers : Dict[str, (int/slice)]
        A dictionary of dim:index where index is the value or range to index the dimension

    coords_array: xarray
        xarray that holds the min/max values of physical coordinates per tile (Round, Ch, Z).

    Returns
    -------
    A coordinates xarray with values recalculated according to indexing on X/Y

    """
    for _round in range(coords_array.sizes[Indices.ROUND]):
        for ch in range(coords_array.sizes[Indices.CH]):
            for z in range(coords_array.sizes[Indices.Z]):
                indices = {
                    Indices.ROUND: _round,
                    Indices.CH: ch,
                    Indices.Z: z
                }
                xmin, xmax, ymin, ymax = coords_array[indices].loc[
                    dict(physical_coordinate=[
                        PhysicalCoordinateTypes.X_MIN,
                        PhysicalCoordinateTypes.X_MAX,
                        PhysicalCoordinateTypes.Y_MIN,
                        PhysicalCoordinateTypes.Y_MAX
                    ])]
                xmin, xmax = _recalculate_physical_coordinate_range(
                    xmin, xmax,
                    stack_shape[Indices.X.value],
                    indexers[Indices.X.value])
                ymin, ymax = _recalculate_physical_coordinate_range(
                    ymin, ymax,
                    stack_shape[Indices.Y.value],
                    indexers[Indices.Y.value])
                coords_array[indices].loc[
                    dict(physical_coordinate=[
                        PhysicalCoordinateTypes.X_MIN,
                        PhysicalCoordinateTypes.X_MAX,
                        PhysicalCoordinateTypes.Y_MIN,
                        PhysicalCoordinateTypes.Y_MAX
                    ])] = [xmin, xmax, ymin, ymax]


def _calculate_physical_pixel_size(coord_min: Number, coord_max: Number, num_pixels: int) -> Number:
    """Calculate the size of a pixel in physical space"""
    return (coord_max - coord_min) / num_pixels


def _pixel_offset_to_physical_coordinate(
        physical_pixel_size: Number,
        pixel_offset: Optional[int],
        coordinates_at_pixel_offset_0: Number,
        dimension_size: int,
) -> Number:
    """Calculate the physical pixel value at the given index"""
    if pixel_offset:
        # Check for negative index
        if pixel_offset < 0:
            pixel_offset = pixel_offset + dimension_size
        return (physical_pixel_size * pixel_offset) + coordinates_at_pixel_offset_0
    return coordinates_at_pixel_offset_0


def _recalculate_physical_coordinate_range(
        coord_min: float,
        coord_max: float,
        dimension_size: int,
        indexer: Union[int, slice],
) -> Tuple[Number, Number]:
    """Given the dimension size and pixel coordinate indexes calculate the corresponding
    coordinates in physical space

    Parameters
    ----------
    coord_min: float
        the minimum physical coordinate value

    coord_max: float
        the maximum physical coordinate value

    dimension_size: int
        The size (number of pixels) to use to calculate physical_pixel_size

    indexer: (int/slice)
        The pixel index or range to calculate.

    Returns
    -------
    The new min and max physical coordinate values of the given dimension
    """
    physical_pixel_size = _calculate_physical_pixel_size(coord_min, coord_max, dimension_size)
    min_pixel_index = indexer if isinstance(indexer, int) else indexer.start
    max_pixel_index = indexer if isinstance(indexer, int) else indexer.stop
    # Add one to max pixel index to get end of pixel
    max_pixel_index = max_pixel_index + 1 if max_pixel_index else dimension_size
    new_min = _pixel_offset_to_physical_coordinate(
        physical_pixel_size, min_pixel_index, coord_min, dimension_size)
    new_max = _pixel_offset_to_physical_coordinate(
        physical_pixel_size, max_pixel_index, coord_min, dimension_size)
    return new_min, new_max


def get_coordinates(
        coords_array: xr.DataArray,
        indices: Mapping[Indices, int],
        physical_axis: Coordinates,
) -> Tuple[float, float]:

    """Given a set of indices that uniquely identify a tile and a physical axis, return the min
    and the max coordinates for that tile along that axis."""

    selectors: Mapping[str, Any] = {
        Indices.ROUND.value: indices[Indices.ROUND],
        Indices.CH.value: indices[Indices.CH],
        Indices.Z.value: indices[Indices.Z],
    }
    min_selectors = dict(selectors)
    max_selectors = dict(selectors)
    if physical_axis == Coordinates.X:
        min_selectors[PHYSICAL_COORDINATE_DIMENSION] = PhysicalCoordinateTypes.X_MIN
        max_selectors[PHYSICAL_COORDINATE_DIMENSION] = PhysicalCoordinateTypes.X_MAX
    elif physical_axis == Coordinates.Y:
        min_selectors[PHYSICAL_COORDINATE_DIMENSION] = PhysicalCoordinateTypes.Y_MIN
        max_selectors[PHYSICAL_COORDINATE_DIMENSION] = PhysicalCoordinateTypes.Y_MAX
    elif physical_axis == Coordinates.Z:
        min_selectors[PHYSICAL_COORDINATE_DIMENSION] = PhysicalCoordinateTypes.Z_MIN
        max_selectors[PHYSICAL_COORDINATE_DIMENSION] = PhysicalCoordinateTypes.Z_MAX

    return (
        coords_array.loc[min_selectors].item(),
        coords_array.loc[max_selectors].item(),
    )


def get_physical_coordinates_of_spot(
        coords_array: xr.DataArray,
        tile_indices: Mapping[Indices, int],
        pixel_x: int,
        pixel_y: int,
        tile_shape: Tuple[int, int]):
    """Given a set of indices that uniquely identify a tile and the location of a spot in pixel space
    calculate the location in physical space."""
    x_range = get_coordinates(coords_array, tile_indices, Coordinates.X)
    physcial_pixel_size_x = _calculate_physical_pixel_size(coord_max=x_range[1],
                                                           coord_min=x_range[0],
                                                           num_pixels=tile_shape[1])
    physical_x = _pixel_offset_to_physical_coordinate(physical_pixel_size=physcial_pixel_size_x,
                                                      pixel_offset=pixel_x,
                                                      coordinates_at_pixel_offset_0=x_range[0],
                                                      dimension_size=tile_shape[1])

    y_range = get_coordinates(coords_array, tile_indices, Coordinates.Y)
    physcial_pixel_size_y = _calculate_physical_pixel_size(coord_max=y_range[1],
                                                           coord_min=y_range[0],
                                                           num_pixels=tile_shape[0])
    physical_y = _pixel_offset_to_physical_coordinate(physical_pixel_size=physcial_pixel_size_y,
                                                      pixel_offset=pixel_y,
                                                      coordinates_at_pixel_offset_0=y_range[0],
                                                      dimension_size=tile_shape[0])

    z_range = get_coordinates(coords_array, tile_indices, Coordinates.Z)
    # As discussed just taking the middle of the z range for this...unless we change our minds
    physical_z = (z_range[1] - z_range[0]) / 2 + z_range[0]

    return physical_x, physical_y, physical_z



from itertools import product
from typing import Iterator, Mapping, Sequence, Tuple

from starfish.types import Indices


def join_dimension_sizes(
        dimension_order: Sequence[Indices],
        *,
        size_for_round: int,
        size_for_ch: int,
        size_for_z: int,
) -> Sequence[Tuple[Indices, int]]:
    """
    Given a sequence of dimensions and their sizes, return a sequence of tuples of dimensions and
    its respective size.

    For example, if dimension_sequence is (ROUND, CH, Z) and each dimension is of size 2, return
    ((ROUND, 2), (CH, 2), (Z, 2)).
    """
    dimension_mapping = {
        Indices.ROUND: size_for_round,
        Indices.CH: size_for_ch,
        Indices.Z: size_for_z,
    }

    return [
        (dimension, dimension_mapping[dimension])
        for dimension in dimension_order
    ]


def ordered_iterator(
        dimension_sizes: Sequence[Tuple[Indices, int]]
) -> Iterator[Mapping[Indices, int]]:
    """
    Given a sequence of tuples of dimensions and its respective size, return an iterator that steps
    through all the possible points in the space.  The sequence is ordered from the slowest varying
    dimension to the fastest varying dimension.
    """
    for tpl in product(*[range(dimension_sizes) for _, dimension_sizes in dimension_sizes]):
        yield dict(zip((index for index, _ in dimension_sizes), tpl))



import pytest
from pkg_resources import resource_filename

from .util import SpaceTxValidator

package_name = "sptx_format"
schema_path = resource_filename(package_name, "schema/fov_manifest.json")
validator = SpaceTxValidator(schema_path)


def test_fov_manifest():
    fov_manifest_example_path = resource_filename(
        package_name, "examples/fov_manifest/fov_manifest.json")
    assert validator.validate_file(fov_manifest_example_path)


def test_empty_manifest_raises_validation_error():
    with pytest.warns(UserWarning):
        assert not validator.validate_object({})



"""
Tests for codebook.from_code_array method
"""

from typing import Any, Dict, List

import numpy as np
import pytest

from starfish import Codebook
from starfish.types import Features, Indices


def codebook_array_factory() -> List[Dict[str, Any]]:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    return [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 2, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]


def test_from_code_array_has_three_channels_two_rounds_and_two_codes():
    """
    Tests that from_code_array loads a small codebook that has the correct size and values
    """
    code_array: List = codebook_array_factory()
    codebook: Codebook = Codebook.from_code_array(code_array)

    assert codebook.sizes[Indices.CH] == 3
    assert codebook.sizes[Indices.ROUND] == 2
    assert codebook.sizes[Features.TARGET] == 2

    # codebook should have 4 "on" combinations
    expected_values = np.zeros((2, 3, 2))
    expected_values[0, 0, 0] = 1
    expected_values[0, 1, 1] = 1
    expected_values[1, 2, 0] = 1
    expected_values[1, 1, 1] = 1

    assert np.array_equal(codebook.values, expected_values)


# TODO ambrosejcarr: this should be a ValueError, not a KeyError,
# and the message should be clearer to the user
def test_from_code_array_throws_key_error_with_missing_channel_round_or_value():
    """Tests that from_code_array throws errors when it encounters malformed codes"""
    code_array: List = codebook_array_factory()

    # codebook is now missing a channel
    del code_array[0][Features.CODEWORD][0][Indices.ROUND.value]
    with pytest.raises(KeyError):
        Codebook.from_code_array(code_array)

    code_array: List = codebook_array_factory()
    del code_array[0][Features.CODEWORD][0][Indices.CH.value]
    with pytest.raises(KeyError):
        Codebook.from_code_array(code_array)

    code_array: List = codebook_array_factory()
    del code_array[0][Features.CODEWORD][0][Features.CODE_VALUE]
    with pytest.raises(KeyError):
        Codebook.from_code_array(code_array)


def test_from_code_array_expands_codebook_when_provided_n_codes_that_exceeds_array_value():
    """
    The codebook factory produces codes with 3 channels and 2 rounds. This test provides numbers
    larger than that, and the codebook should be expanded to those numbers as a result.
    """
    code_array: List = codebook_array_factory()
    codebook: Codebook = Codebook.from_code_array(code_array, n_ch=10, n_round=4)
    assert codebook.sizes[Indices.CH] == 10
    assert codebook.sizes[Indices.ROUND] == 4
    assert codebook.sizes[Features.TARGET] == 2


def test_from_code_array_throws_exceptions_when_data_does_not_match_channel_or_round_requests():
    """
    The codebook factory produces codes with 3 channels and 2 rounds. This test provides numbers
    larger than that, and the codebook should be expanded to those numbers as a result.
    """
    code_array: List = codebook_array_factory()

    # should throw an exception, as 3 channels are present in the data
    with pytest.raises(ValueError):
        Codebook.from_code_array(code_array, n_ch=2, n_round=4)

    # should throw an exception, as 2 rounds are present in the data
    with pytest.raises(ValueError):
        Codebook.from_code_array(code_array, n_ch=3, n_round=1)


def test_from_code_array_throws_exception_when_data_is_improperly_formatted():
    code_array: List = codebook_array_factory()
    code_array[0][Features.CODEWORD][0] = ('I should be a dict, oops!',)
    with pytest.raises(TypeError):
        Codebook.from_code_array(code_array, n_ch=3, n_round=1)


# TODO codebook should throw an error when an empty array is passed



import copy
import json
import os
import sys
import warnings
from typing import Any, Dict, IO, Iterator, List, Optional, Union

from jsonschema import Draft4Validator, RefResolver, ValidationError
from pkg_resources import resource_filename

package_name = 'sptx_format'


class SpaceTxValidator:

    def __init__(self, schema: str) -> None:
        """create a validator for a json-schema compliant spaceTx specification file

        Parameters
        ----------
        schema : str
            file path to schema

        """
        self._schema: Dict = self.load_json(schema)
        self._validator: Draft4Validator = self._create_validator(self._schema)

    @staticmethod
    def _create_validator(schema: Dict) -> Draft4Validator:
        """resolve $ref links in a loaded json schema and return a validator

        Parameters
        ----------
        schema : Dict
            loaded json schema

        Returns
        -------
        Draft4Validator :
            json-schema validator specific to the supplied schema, with references resolved

        """
        experiment_schema_path = resource_filename(package_name, "schema/experiment.json")
        package_root = os.path.dirname(os.path.dirname(experiment_schema_path))
        base_uri = 'file://' + package_root + '/'
        resolver = RefResolver(base_uri, schema)
        return Draft4Validator(schema, resolver=resolver)

    @staticmethod
    def load_json(json_file: str) -> Dict:
        with open(json_file, 'rb') as f:
            return json.load(f)

    @staticmethod
    def _recurse_through_errors(error_iterator: Iterator[ValidationError],
                                level: int=0,
                                filename: str=None) -> None:
        """Recurse through ValidationErrors, printing message and schema path

        Parameters
        ----------
        error_iterator : Iterator[ValidationError]
            iterator over ValidationErrors that occur during validation
        level : int
            current level of recursion
        filename: str
            informational string regarding the source file of the given object

        """
        fmt = "\n{stars} {message}\n"
        fmt += "\tSchema:         \t{schema}\n"
        fmt += "\tSubschema level:\t{level}\n"
        fmt += "\tPath to error:  \t{path}\n"
        if filename:
            fmt += "\tFilename:       \t{filename}\n"
        for error in error_iterator:
            message = fmt.format(
                stars="***" * level, level=str(level), path="/".join(error.absolute_schema_path),
                message=error.message, cause=error.cause, schema=error.schema.get("$id", "unknown"),
                filename=filename,
            )
            warnings.warn(message)
            if error.context:
                level += 1
                SpaceTxValidator._recurse_through_errors(error.context, level=level)

    def validate_file(self, target_file: str) -> bool:
        """validate a target file, returning True if valid and False otherwise

        Parameters
        ----------
        target_file : str
            path or URL to a target json object to be validated against the schema passed to this
            object's constructor

        Returns
        -------
        bool :
            True, if object valid, else False

        """
        target_object = self.load_json(target_file)
        return self.validate_object(target_object, target_file)

    def validate_object(
            self,
            target_object: Union[dict, list],
            target_file: str=None,
    ) -> bool:
        """validate a loaded json object, returning True if valid, and False otherwise

        Parameters
        ----------
        target_object : Dict
            loaded json object to be validated against the schema passed to this object's
            constructor
        target_file : str
            informational string regarding the source file of the given object

        Returns
        -------
        bool :
            True, if object valid, else False

        """

        if self._validator.is_valid(target_object):
            return True
        else:
            es: Iterator[ValidationError] = self._validator.iter_errors(target_object)
            self._recurse_through_errors(es, filename=target_file)
            return False

    def fuzz_object(
            self,
            target_object: Union[dict, list],
            target_file: str=None,
            out: IO=sys.stdout,
    ) -> None:
        """performs mutations on the given object and tests for validity.

        A representation of the validity is printed to the given output stream.

        Parameters
        ----------
        target_object : Dict
            loaded json object to be validated against the schema passed to this object's
            constructor
        target_file : str
            informational string regarding the source file of the given object
        out : IO
            output stream for printing

        """

        if target_file:
            out.write(f"> Fuzzing {target_file}...\n")
        else:
            out.write("> Fuzzing unknown...\n")
        fuzzer = Fuzzer(self._validator, target_object, out)
        fuzzer.fuzz()


class Fuzzer(object):

    def __init__(self, validator: Draft4Validator, obj: Any, out: IO=sys.stdout) -> None:
        """create a fuzzer which will check different situations against the validator

        Parameters
        ----------
        validator : SpaceTxValidator
            validator which should match the given object type
        obj : Any
            JSON-like object which will be checked against the validator
        out : IO
            output stream for printing

        """
        self.validator = validator
        self.obj = obj
        self.out = out
        self.stack: Optional[List[Any]] = None

    def fuzz(self) -> None:
        """prints to the out field the state of the object tree after types of fuzzing

        Each line is prefixed by the output of {state()} followed by a YAML-like
        representation of the branch of the object tree.
        """
        header = f"{self.state()}"
        header += "If the letter is present, mutation is valid!"
        self.out.write(f"{header}\n")
        self.out.write("".join([x in ("\t", "\n") and x or "-" for x in header]))
        self.out.write("\n")
        self.stack: List[Any] = []
        try:
            self.descend(self.obj)
        finally:
            self.stack = None

    def state(self) -> str:
        """primary driver for the checks of individual trees

        Returns
        -------
        str :
            space-separated representation of the fuzzing conditions.
            If a letter is present, then mutation leaves the tree in
            a valid state:

             A: inserting a fake key or appending to a list
             D: deleting a key or index
             I: converting value to an integer
             I: converting value to a string
             M: converting value to an empty dict
             L: converting value to an empty list

        """
        rv = [
            Add().check(self),
            Del().check(self),
            Change("I", lambda *args: 123456789).check(self),
            Change("S", lambda *args: "fake").check(self),
            Change("M", lambda *args: dict()).check(self),
            Change("L", lambda *args: list()).check(self),
        ]
        return ' '.join(rv) + "\t"

    def descend(self, obj: Any, depth: int=0, prefix: str="") -> None:
        """walk a JSON-like object tree printing the state of the tree
        at each level. A YAML representation is used for simplicity.

        Parameters
        ----------
        obj : Any
            JSON-like object tree
        depth : int
            depth in the tree that is currently being evaluated
        prefix : str
            value which should be prepended to printouts at this level
        """
        if self.stack is None:
            raise Exception("invalid state")
        if isinstance(obj, list):
            for i, o in enumerate(obj):
                depth += 1
                self.stack.append(i)
                self.descend(o, depth, prefix="- ")
                self.stack.pop()
                depth -= 1
        elif isinstance(obj, dict):
            for k in obj:
                # This is something of a workaround in that we need a special
                # case for object keys since no __getitem__ method will suffice.
                self.stack.append((k,))
                self.out.write(f"{self.state()}{' ' * depth}{prefix}{k}:\n")
                self.stack.pop()
                if prefix == "- ":
                    prefix = "  "
                depth += 1
                self.stack.append(k)
                self.descend(obj[k], depth, prefix="  " + prefix)
                self.stack.pop()
                depth -= 1
        else:
            self.out.write(f"{self.state()}{' ' * depth}{prefix}{obj}\n")

class Checker(object):

    @property
    def LETTER(self) -> str:
        return "?"

    def check(self, fuzz: Fuzzer) -> str:
        """create a copy of the current state of the object tree,
        mutate it, and run it through is_valid on the validator.

        Parameters
        ----------
        fuzz : Fuzzer
            the containing instance

        Returns
        -------
        str :
            A single character string representation of the check

        """
        # Don't mess with the top level
        if fuzz.stack is None:
            return self.LETTER
        if not fuzz.stack:
            return "-"
        # Operate on a copy for mutating
        dupe = copy.deepcopy(fuzz.obj)
        target = dupe
        for level in fuzz.stack[0:-1]:
            target = target.__getitem__(level)
        self.handle(fuzz, target)
        valid = fuzz.validator.is_valid(dupe)
        return valid and self.LETTER or "."

    def handle(self, fuzz, target):
        raise NotImplementedError()

class Add(Checker):

    @property
    def LETTER(self) -> str:
        return "A"

    def handle(self, fuzz, target):
        if isinstance(target, dict):
            target["fake"] = "!"
        elif isinstance(target, list):
            target.append("!")
        else:
            raise Exception("unknown")

class Del(Checker):

    @property
    def LETTER(self) -> str:
        return "D"

    def handle(self, fuzz, target):
        key = fuzz.stack[-1]
        if isinstance(key, tuple):
            key = key[0]
        target.__delitem__(key)

class Change(Checker):

    @property
    def LETTER(self) -> str:
        return self.letter

    def __init__(self, letter, call):
        self.letter = letter
        self.call = call

    def handle(self, fuzz, target):
        key = fuzz.stack[-1]
        if isinstance(key, tuple):
            key = key[0]
        target.__setitem__(key, self.call())



from collections import OrderedDict

import numpy as np

from starfish import Codebook, ImageStack
from starfish.types import Features, Indices, PhysicalCoordinateTypes


def imagestack_with_coords_factory(stack_shape: OrderedDict, coords: OrderedDict) -> ImageStack:
    """
    Create an ImageStack of given shape and assigns the given x,y,z
    min/max physical coordinates to each tile.

    Parameters
    ----------
    stack_shape: OrderedDict
        Dict[Indices, int] defining the size of each dimension for an ImageStack

    coords: OrderedDict
        Dict[PhysicalCoordinateTypes, float] defining the min/max values of physical
        coordinates to assign to each tile of the return ImageStack
    """

    stack = ImageStack.synthetic_stack(num_round=stack_shape[Indices.ROUND],
                                       num_ch=stack_shape[Indices.CH],
                                       num_z=stack_shape[Indices.Z],
                                       tile_height=stack_shape[Indices.Y],
                                       tile_width=stack_shape[Indices.X])

    coords_array = [coords[PhysicalCoordinateTypes.X_MIN],
                    coords[PhysicalCoordinateTypes.X_MAX],
                    coords[PhysicalCoordinateTypes.Y_MIN],
                    coords[PhysicalCoordinateTypes.Y_MAX],
                    coords[PhysicalCoordinateTypes.Z_MIN],
                    coords[PhysicalCoordinateTypes.Z_MAX]]

    for _round in range(stack.num_rounds):
        for ch in range(stack.num_chs):
            for z in range(stack.num_zlayers):
                coordinate_selector = {
                    Indices.ROUND.value: _round,
                    Indices.CH.value: ch,
                    Indices.Z.value: z,
                }

                stack._coordinates.loc[coordinate_selector] = np.array(coords_array)

    return stack


def codebook_array_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    data = [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 2, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    return Codebook.from_code_array(data)



import os
import sys
import unittest

from starfish.test.full_pipelines.cli._base_cli_test import CLITest


@unittest.skip("skipping for now")
class TestAllenData(CLITest, unittest.TestCase):

    @property
    def subdirs(self):
        return (
            "registered",
            "filtered",
            "results"
        )

    @property
    def stages(self):
        return (
            [
                sys.executable,
                "examples/get_cli_test_data.py",
                "https://d2nhj9g34unfro.cloudfront.net/20180828/"
                + "allen_smFISH-TEST/allen_smFISH_test_data.zip",
                lambda tempdir, *args, **kwargs: os.path.join(tempdir, "registered")
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered/fov_001", "hybridization.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "clip_filtered.json"),
                "Clip",
                "--p-min", "10",
                "--p-max", "100"
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "clip_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "bandpass_filtered.json"),
                "Bandpass",
                "--lshort", ".5",
                "--llong", "7",
                "--truncate", "4"
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "bandpass_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "clip2_filtered.json"),
                "Clip",
                "--p-min", "10",
                "--p-max", "100"
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "clip2_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "gaussian_filtered.json"),
                "GaussianLowPass",
                "--sigma", "1"
            ],
            [
                "starfish", "detect_spots",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "gaussian_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results"),
                "LocalMaxPeakFinder",
                "--spot-diameter", "3",
                "--min-mass", "300",
                "--max-size", "3",
                "--separation", "5",
                "--percentile", "10",
                "--is-volume"
            ],

        )

    def verify_results(self, intensities):
        # TODO DEEP SAYS WAIT ON THIS TEST TILL STUFF GETS FIGURED OUT
        pass



#!/usr/bin/env python
import cProfile
from pstats import Stats

import click

from sptx_format.cli import validate as validate_cli
from starfish.experiment.builder.cli import build as build_cli
from starfish.image import (
    Filter,
    Registration,
    Segmentation,
)
from starfish.spots import (
    Decoder,
    SpotFinder,
    TargetAssignment,
)


PROFILER_KEY = "profiler"
"""This is the dictionary key we use to attach the profiler to pass to the resultcallback."""
PROFILER_LINES = 15
"""This is the number of profiling rows to dump when --profile is enabled."""


@click.group()
@click.option("--profile", is_flag=True)
@click.pass_context
def starfish(ctx, profile):
    art = """
         _              __ _     _
        | |            / _(_)   | |
     ___| |_ __ _ _ __| |_ _ ___| |__
    / __| __/ _` | '__|  _| / __| '_  `
    \__ \ || (_| | |  | | | \__ \ | | |
    |___/\__\__,_|_|  |_| |_|___/_| |_|

    """  # noqa
    print_art = True
    sub = ctx.command.get_command(ctx, ctx.invoked_subcommand)
    if hasattr(sub, "no_art"):
        print_art = not sub.no_art
    if print_art:
        print(art)
    if profile:
        profiler = cProfile.Profile()
        profiler.enable()

        def print_profile():
            stats = Stats(profiler)
            stats.sort_stats('tottime').print_stats(PROFILER_LINES)

        ctx.call_on_close(print_profile)


@starfish.command()
def version():
    import pkg_resources
    version = pkg_resources.require("starfish")[0].version
    print(version)
version.no_art = True  # type: ignore


# Pipelines
starfish.add_command(Registration._cli)  # type: ignore
starfish.add_command(Filter._cli)  # type: ignore
starfish.add_command(SpotFinder._cli)  # type: ignore
starfish.add_command(Segmentation._cli)  # type: ignore
starfish.add_command(TargetAssignment._cli)  # type: ignore
starfish.add_command(Decoder._cli)  # type: ignore

# Other
starfish.add_command(build_cli)  # type: ignore
starfish.add_command(validate_cli)  # type: ignore



import pytest

from starfish.experiment.experiment import Experiment
from starfish.experiment.version import MAX_SUPPORTED_VERSION, MIN_SUPPORTED_VERSION


def test_min_version():
    with pytest.raises(ValueError):
        Experiment.verify_version("0.0.0-dev")
    Experiment.verify_version(str(MIN_SUPPORTED_VERSION))
    Experiment.verify_version(str(MAX_SUPPORTED_VERSION))


def test_max_version():
    with pytest.raises(ValueError):
        Experiment.verify_version("4294967296.0.0")
    Experiment.verify_version(str(MIN_SUPPORTED_VERSION))
    Experiment.verify_version(str(MAX_SUPPORTED_VERSION))



import numpy as np

from starfish import ImageStack
from starfish.image._filter.zero_by_channel_magnitude import ZeroByChannelMagnitude

def create_imagestack_with_magnitude_scale():
    """create an imagestack with increasing magnitudes"""
    data = np.linspace(0, 1, 11, dtype=np.float32)
    data = np.repeat(data[None, :], 2, axis=0)
    # reshape data into a 2-channel, (1, 11, 1) image in (x, y, z)
    data = data.reshape(1, 2, 1, 11, 1)
    imagestack = ImageStack.from_numpy_array(data)
    return imagestack


def test_zero_by_channel_magnitude_produces_accurate_results():
    imagestack = create_imagestack_with_magnitude_scale()

    zcm = ZeroByChannelMagnitude(thresh=np.inf, normalize=False)
    filtered = zcm.run(imagestack, in_place=False, n_processes=1)
    assert np.all(filtered.xarray == 0)



import os
import shutil

from starfish.intensity_table.intensity_table import IntensityTable
from starfish.util import exec


class CLITest:
    """This is a base class for testing CLI methods. Each stage should correspond
    to a different pipeline step. Running the test will go through each stage and
    run the command line method with given arguments. The last stage should produce
    a file called results containing an IntensityTable. Each cli test should define
    it's own verify_results method.
    """
    @property
    def subdirs(self):
        raise NotImplementedError()

    @property
    def stages(self):
        raise NotImplementedError()

    @property
    def spots_file(self):
        return "spots.nc"

    def verify_results(self, intensities: IntensityTable):
        raise NotImplementedError()

    def test_run_pipline(self):
        tempdir = exec.stages(self.stages, self.subdirs, keep_data=True)
        intensities = IntensityTable.load(os.path.join(tempdir, "results", self.spots_file))
        self.verify_results(intensities)

        if os.getenv("TEST_KEEP_DATA") is None:
            shutil.rmtree(tempdir)



"""
Tests for codebook.synthetic_one_hot_codebook method
"""

import numpy as np
import pytest

from starfish import Codebook
from starfish.types import Features, Indices


def test_synthetic_one_hot_codebook_returns_requested_codebook():
    """
    Make a request and verify that the size and shape match the request, and
    that each round has only one round 'on'.
    """
    codebook: Codebook = Codebook.synthetic_one_hot_codebook(
        n_round=4,
        n_channel=2,
        n_codes=3
    )

    assert codebook.sizes == {Indices.CH: 2, Indices.ROUND: 4, Features.TARGET: 3}
    assert np.all(codebook.sum(Indices.CH.value) == 1), "the numbers of channels on per round != 1"


def test_target_names_are_incorporated_into_synthetic_codebook():
    """Verify that target names are incorporated in order."""
    codebook: Codebook = Codebook.synthetic_one_hot_codebook(
        n_round=3,
        n_channel=6,
        n_codes=2,
        target_names=list('ab'),
    )

    assert np.array_equal(codebook[Features.TARGET], list('ab'))


# TODO ambrosejcarr: This should probably be clearer than an AssertionError
def test_wrong_number_of_target_names_raises_error():
    """here we request 3 codes but provide only two names, which should raise an error"""
    with pytest.raises(AssertionError):
        Codebook.synthetic_one_hot_codebook(
            n_round=2,
            n_channel=2,
            n_codes=3,
            target_names=list('ab')
        )



from typing import Mapping, Tuple, Union

import numpy as np
from slicedimage import ImageFormat

from starfish.experiment.builder import FetchedTile, tile_fetcher_factory
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Coordinates, Indices, Number

NUM_ROUND = 8
NUM_CH = 1
NUM_Z = 1
HEIGHT = 10
WIDTH = 10


def round_to_x(r: int) -> Tuple[float, float]:
    return (r + 1) * 1000, (r + 1) * 100


def round_to_y(r: int) -> Tuple[float, float]:
    return (r + 1) * 10, (r + 1) * 0.1


def round_to_z(r: int) -> Tuple[float, float]:
    return (r + 1) * 0.01, (r + 1) * 0.001


class OffsettedTiles(FetchedTile):
    """Tiles that are physically offset based on round."""
    def __init__(self, fov: int, _round: int, ch: int, z: int) -> None:
        super().__init__()
        self._round = _round

    @property
    def shape(self) -> Tuple[int, ...]:
        return HEIGHT, WIDTH

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        return {
            Coordinates.X: round_to_x(self._round),
            Coordinates.Y: round_to_y(self._round),
            Coordinates.Z: round_to_z(self._round),
        }

    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    def tile_data(self) -> np.ndarray:
        return np.ones((HEIGHT, WIDTH), dtype=np.float32)


def test_coordinates():
    """Set up an ImageStack with tiles that are offset based on round.  Verify that the coordinates
    retrieved match.
    """
    stack = ImageStack.synthetic_stack(
        NUM_ROUND, NUM_CH, NUM_Z,
        HEIGHT, WIDTH,
        tile_fetcher=tile_fetcher_factory(
            OffsettedTiles,
            True,
        )
    )

    for _round in range(NUM_ROUND):
        for ch in range(NUM_CH):
            for z in range(NUM_Z):
                indices = {
                    Indices.ROUND: _round,
                    Indices.CH: ch,
                    Indices.Z: z
                }

                xmin, xmax = stack.coordinates(indices, Coordinates.X)
                ymin, ymax = stack.coordinates(indices, Coordinates.Y)
                zmin, zmax = stack.coordinates(indices, Coordinates.Z)

                expected_xmin, expected_xmax = round_to_x(_round)
                expected_ymin, expected_ymax = round_to_y(_round)
                expected_zmin, expected_zmax = round_to_z(_round)

                assert np.isclose(xmin, expected_xmin)
                assert np.isclose(xmax, expected_xmax)
                assert np.isclose(ymin, expected_ymin)
                assert np.isclose(ymax, expected_ymax)
                assert np.isclose(zmin, expected_zmin)
                assert np.isclose(zmax, expected_zmax)


class OffsettedScalarTiles(FetchedTile):
    """Tiles that are physically offset based on round, but only have a single scalar coordinate."""
    def __init__(self, fov: int, _round: int, ch: int, z: int) -> None:
        super().__init__()
        self._round = _round

    @property
    def shape(self) -> Tuple[int, ...]:
        return HEIGHT, WIDTH

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        return {
            Coordinates.X: round_to_x(self._round)[0],
            Coordinates.Y: round_to_y(self._round)[0],
            Coordinates.Z: round_to_z(self._round)[0],
        }

    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    def tile_data(self) -> np.ndarray:
        return np.ones((HEIGHT, WIDTH), dtype=np.float32)


def test_scalar_coordinates():
    """Set up an ImageStack where only a single scalar physical coordinate is provided per axis.
    Internally, this should be converted to a range where the two endpoints are identical to the
    physical coordinate provided.
    """
    stack = ImageStack.synthetic_stack(
        NUM_ROUND, NUM_CH, NUM_Z,
        HEIGHT, WIDTH,
        tile_fetcher=tile_fetcher_factory(
            OffsettedScalarTiles,
            True,
        )
    )

    for _round in range(NUM_ROUND):
        for ch in range(NUM_CH):
            for z in range(NUM_Z):
                indices = {
                    Indices.ROUND: _round,
                    Indices.CH: ch,
                    Indices.Z: z
                }

                xmin, xmax = stack.coordinates(indices, Coordinates.X)
                ymin, ymax = stack.coordinates(indices, Coordinates.Y)
                zmin, zmax = stack.coordinates(indices, Coordinates.Z)

                expected_x = round_to_x(_round)[0]
                expected_y = round_to_y(_round)[0]
                expected_z = round_to_z(_round)[0]

                assert np.isclose(xmin, expected_x)
                assert np.isclose(xmax, expected_x)
                assert np.isclose(ymin, expected_y)
                assert np.isclose(ymax, expected_y)
                assert np.isclose(zmin, expected_z)
                assert np.isclose(zmax, expected_z)



import json
import uuid
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from semantic_version import Version
from sklearn.neighbors import NearestNeighbors
from slicedimage.io import resolve_path_or_url

from starfish.codebook._format import (
    CURRENT_VERSION,
    DocumentKeys,
    MAX_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION,
)
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Features, Indices, Number


class Codebook(xr.DataArray):
    """Codebook for an image-based transcriptomics experiment

    The codebook is a three dimensional tensor whose values are the expected intensity of a spot
    for each code in each imaging round and each color channel. This class supports the
    construction of synthetic codebooks for testing, and exposes decode methods to assign target
    identifiers to spots. This codebook provides an in-memory representation of the codebook
    defined in the spaceTx format.

    The codebook is a subclass of xarray, and exposes the complete public API of that package in
    addition to the methods and constructors listed below.

    Methods
    -------
    from_code_array(code_array, n_round, n_ch)
        construct a codebook from a spaceTx-spec array of codewords

    from_json(json_codebook, n_round, n_ch)
        load a codebook from a spaceTx spec-compliant json file

    to_json(filename)
        save a codebook to json

    synthetic_one_hot_codebook(n_round, n_channel, n_codes, target_names=None)
        Construct a codebook of random codes where only one channel is on per imaging round.
        This is the typical codebook format for in-situ sequencing and non-multiplex smFISH
        experiments.

    decode_euclidean(intensities)
        find the closest code for each spot in intensities by euclidean distance

    decode_per_round_maximum(intensities)
        find codes that match the per-channel max intensity for each spot in intensities

    code_length()
        return the total length of the codes in the codebook

    Examples
    --------
    Build a codebook using ``Codebook.synthetic_one_hot_codebook``::

        >>> from starfish import Codebook
        >>> sd = Codebook.synthetic_one_hot_codebook(n_ch=3, n_round=4, n_codes=2)
        >>> sd.codebook()
        <xarray.Codebook (target: 2, c: 3, h: 4)>
        array([[[0, 0, 0, 0],
                [0, 0, 1, 1],
                [1, 1, 0, 0]],

               [[1, 0, 0, 0],
                [0, 0, 1, 0],
                [0, 1, 0, 1]]], dtype=uint8)
        Coordinates:
          * target     (target) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...
          * c          (c) int64 0 1 2
          * h          (h) int64 0 1 2 3

    See Also
    --------
    TODO <link to spaceTx format>

    """

    @property
    def code_length(self) -> int:
        """return the length of codes in this codebook"""
        return int(np.dot(*self.shape[1:]))

    @classmethod
    def _empty_codebook(cls, code_names: Sequence[str], n_ch: int, n_round: int):
        """create an empty codebook of shape (code_names, n_ch, n_round)

        Parameters
        ----------
        code_names : Sequence[str]
            the targets to be coded
        n_ch : int
            number of channels used to build the codes
        n_round : int
            number of imaging rounds used to build the codes

        Examples
        --------
        >>> from starfish import Codebook
        >>> Codebook._empty_codebook(['ACTA', 'ACTB'], n_ch=3, n_round=2)
        <xarray.Codebook (target: 2, c: 3, h: 2)>
        array([[[0, 0],
                [0, 0],
                [0, 0]],

               [[0, 0],
                [0, 0],
                [0, 0]]], dtype=uint8)
        Coordinates:
          * target     (target) object 'ACTA' 'ACTB'
          * c          (c) int64 0 1 2
          * h          (h) int64 0 1

        Returns
        -------
        Codebook :
            codebook whose values are all zero

        """
        codes_index = pd.Index(code_names, name=Features.TARGET)
        return cls(
            data=np.zeros((codes_index.shape[0], n_ch, n_round), dtype=np.uint8),
            coords=(
                codes_index,
                pd.Index(np.arange(n_ch), name=Indices.CH.value),
                pd.Index(np.arange(n_round), name=Indices.ROUND.value),
            )
        )

    @classmethod
    def _verify_version(cls, semantic_version_str: str) -> None:
        version = Version(semantic_version_str)
        if not (MIN_SUPPORTED_VERSION <= version <= MAX_SUPPORTED_VERSION):
            raise ValueError(
                f"version {version} not supported.  This version of the starfish library only "
                f"supports codebook formats from {MIN_SUPPORTED_VERSION} to "
                f"{MAX_SUPPORTED_VERSION}")

    @classmethod
    def from_code_array(
            cls, code_array: List[Dict[Union[str, Any], Any]],
            n_round: Optional[int]=None, n_ch: Optional[int]=None) -> "Codebook":
        """construct a codebook from a spaceTx-spec array of codewords

        Parameters
        ----------
        code_array : List[Dict[str, Any]]
            Array of dictionaries, each containing a codeword and target
        n_round : Optional[int]
            The number of imaging rounds used in the codes. Will be inferred if not provided
        n_ch : Optional[int]
            The number of channels used in the codes. Will be inferred if not provided

        Examples
        --------
        Construct a codebook from some array data in python memory::

            >>> from starfish.types import Indices
            >>> from starfish import Codebook
            >>> codebook = [
            >>>     {
            >>>         Features.CODEWORD: [
            >>>             {Indices.ROUND.value: 0, Indices.CH.value: 3, Features.CODE_VALUE: 1},
            >>>             {Indices.ROUND.value: 1, Indices.CH.value: 3, Features.CODE_VALUE: 1},
            >>>         ],
            >>>         Features.TARGET: "ACTB_human"
            >>>     },
            >>>     {
            >>>         Features.CODEWORD: [
            >>>             {Indices.ROUND.value: 0, Indices.CH.value: 3, Features.CODE_VALUE: 1},
            >>>             {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1},
            >>>         ],
            >>>         Features.TARGET: "ACTB_mouse"
            >>>     },
            >>> ]
            >>> Codebook.from_code_array(codebook)
            <xarray.Codebook (target: 2, c: 4, h: 2)>
            array([[[0, 0],
                    [0, 0],
                    [0, 0],
                    [1, 1]],

                   [[0, 0],
                    [0, 1],
                    [0, 0],
                    [1, 0]]], dtype=uint8)
            Coordinates:
              * target     (target) object 'ACTB_human' 'ACTB_mouse'
              * c          (c) int64 0 1 2 3
              * h          (h) int64 0 1

            Codebook.from_json(json_codebook)

            Returns
            -------
            Codebook :
                Codebook with shape (targets, channels, imaging_rounds)

        """

        # guess the max round and channel if not provided, otherwise check provided values are valid
        max_round, max_ch = 0, 0

        for code in code_array:
            for entry in code[Features.CODEWORD]:
                max_round = max(max_round, entry[Indices.ROUND])
                max_ch = max(max_ch, entry[Indices.CH])

        # set n_ch and n_round if either were not provided
        n_round = n_round if n_round is not None else max_round + 1
        n_ch = n_ch if n_ch is not None else max_ch + 1

        # raise errors if provided n_round or n_ch are out of range
        if max_round + 1 > n_round:
            raise ValueError(
                f'code detected that requires an imaging round value ({max_round + 1}) that is '
                f'greater than provided n_round: {max_round}')
        if max_ch + 1 > n_ch:
            raise ValueError(
                f'code detected that requires a channel value ({max_ch + 1}) that is greater '
                f'than provided n_ch: {n_ch}')

        # verify codebook structure and fields
        for code in code_array:

            if not isinstance(code, dict):
                raise ValueError(f'codebook must be an array of dictionary codes. Found: {code}.')

            # verify all necessary fields are present
            required_fields = {Features.CODEWORD, Features.TARGET}
            missing_fields = required_fields.difference(code)
            if missing_fields:
                raise ValueError(
                    f'Each entry of codebook must contain {required_fields}. Missing fields: '
                    f'{missing_fields}')

        target_names = [w[Features.TARGET] for w in code_array]
        code_data = cls._empty_codebook(target_names, n_ch, n_round)

        # fill the codebook
        for code_dict in code_array:
            codeword = code_dict[Features.CODEWORD]
            target = code_dict[Features.TARGET]
            for entry in codeword:
                code_data.loc[target, entry[Indices.CH], entry[Indices.ROUND]] = entry[
                    Features.CODE_VALUE]

        return code_data

    @classmethod
    def from_json(
            cls, json_codebook: str, n_round: Optional[int]=None, n_ch: Optional[int]=None
    ) -> "Codebook":
        """Load a codebook from a spaceTx spec-compliant json file or a url pointing to such a file

        Parameters
        ----------
        json_codebook : str
            path or url to json file containing a spaceTx codebook
        n_round : Optional[int]
            The number of imaging rounds used in the codes. Will be inferred if not provided
        n_ch : Optional[int]
            The number of channels used in the codes. Will be inferred if not provided

        Examples
        --------
        Create a codebook from in-memory data::

            >>> from starfish.types import Indices
            >>> from starfish import Codebook
            >>> import tempfile
            >>> import json
            >>> import os
            >>> dir_ = tempfile.mkdtemp()
            >>> codebook = [
            >>>     {
            >>>         Features.CODEWORD: [
            >>>             {Indices.ROUND.value: 0, Indices.CH.value: 3, Features.CODE_VALUE: 1},
            >>>             {Indices.ROUND.value: 1, Indices.CH.value: 3, Features.CODE_VALUE: 1},
            >>>         ],
            >>>         Features.TARGET: "ACTB_human"
            >>>     },
            >>>     {
            >>>         Features.CODEWORD: [
            >>>             {Indices.ROUND.value: 0, Indices.CH.value: 3, Features.CODE_VALUE: 1},
            >>>             {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1},
            >>>         ],
            >>>         Features.TARGET: "ACTB_mouse"
            >>>     },
            >>> ]
            >>> # make a fake file
            >>> json_codebook = os.path.join(dir_, 'codebook.json')
            >>> with open(json_codebook, 'w') as f:
            >>>     json.dump(codebook, f)
            >>> # read codebook from file
            >>> Codebook.from_json(json_codebook)
           <xarray.Codebook (target: 2, c: 4, h: 2)>
            array([[[0, 0],
                    [0, 0],
                    [0, 0],
                    [1, 1]],

                   [[0, 0],
                    [0, 1],
                    [0, 0],
                    [1, 0]]], dtype=uint8)
            Coordinates:
              * target     (target) object 'ACTB_human' 'ACTB_mouse'
              * c          (c) int64 0 1 2 3
              * h          (h) int64 0 1

        Returns
        -------
        Codebook :
            Codebook with shape (targets, channels, imaging_rounds)

        """
        backend, name, _ = resolve_path_or_url(json_codebook)
        with backend.read_contextmanager(name) as fh:
            codebook_doc = json.load(fh)

        if isinstance(codebook_doc, list):
            raise ValueError(
                f"codebook is a list and not an dictionary.  It is highly likely that you are using"
                f"a codebook formatted for a previous version of starfish.")

        version_str = codebook_doc[DocumentKeys.VERSION_KEY]
        cls._verify_version(version_str)

        return cls.from_code_array(codebook_doc[DocumentKeys.MAPPINGS_KEY], n_round, n_ch)

    def to_json(self, filename: str) -> None:
        """save a codebook to json

        Notes
        -----
        This enforces the following typing of codebooks:
        ch, round : int
        value : float
        target : str

        Parameters
        ----------
        filename : str
            filename

        """
        code_array = []
        for target in self[Features.TARGET]:
            codeword = []
            for ch in self[Indices.CH.value]:
                for round_ in self[Indices.ROUND.value]:
                    if self.loc[target, ch, round_]:
                        codeword.append(
                            {
                                Indices.CH.value: int(ch),
                                Indices.ROUND.value: int(round_),
                                Features.CODE_VALUE: float(self.loc[target, ch, round_])
                            })
            code_array.append({
                Features.CODEWORD: codeword,
                Features.TARGET: str(target.values)
            })
        codebook_document = {
            DocumentKeys.VERSION_KEY: str(CURRENT_VERSION),
            DocumentKeys.MAPPINGS_KEY: code_array,
        }

        with open(filename, 'w') as f:
            json.dump(codebook_document, f)

    @staticmethod
    def _normalize_features(
            array: Union["Codebook", IntensityTable],
            norm_order,
    ) -> Tuple[Union["Codebook", IntensityTable], np.ndarray]:
        """unit normalize each feature of array

        Parameters
        ----------
        array : Union[Codebook, IntensityTable]
            codebook or intensity table containing (ch, r) features to normalize
        norm_order : int
            the norm to apply to each feature

        See Also
        --------
        The available norms for this function can be found at the following link:
        https://docs.scipy.org/doc/numpy-1.14.0/reference/generated/numpy.linalg.norm.html

        Returns
        -------
        Union[IntensityTable, Codebook] :
            IntensityTable or Codebook containing normalized features
        np.ndarray :
            A 1 dimensional numpy array containing the feature norms

        """
        feature_traces = array.stack(traces=(Indices.CH.value, Indices.ROUND.value))
        norm = np.linalg.norm(feature_traces.values, ord=norm_order, axis=1)
        array = array / norm[:, None, None]

        # if a feature is all zero, the information should be spread across the channel
        n = array.sizes[Indices.CH.value] * array.sizes[Indices.ROUND.value]
        partitioned_intensity = np.linalg.norm(np.full(n, fill_value=1 / n), ord=norm_order) / n
        array = array.fillna(partitioned_intensity)

        return array, norm

    @staticmethod
    def _approximate_nearest_code(
            norm_codes: "Codebook", norm_intensities: xr.DataArray, metric: str,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """find the nearest code for each feature using the ball_tree approximate NN algorithm

        Parameters
        ----------
        norm_codes : Codebook
            codebook with each code normalized to unit length (sum = 1)
        norm_intensities : IntensityTable
            intensity table with each feature normalized to unit length (sum = 1)
        metric : str
            the sklearn metric string to pass to NearestNeighbors

        Returns
        -------
        np.ndarray : metric_output
            the output of metric applied to each feature closest code
        np.ndarray : targets
            the gene that corresponds to each matched code

        Notes
        -----
        This function does not verify that the intensities have been normalized.

        """
        linear_codes = norm_codes.stack(traces=(Indices.CH.value, Indices.ROUND.value)).values
        linear_features = norm_intensities.stack(
            traces=(Indices.CH.value, Indices.ROUND.value)).values

        # reshape into traces
        nn = NearestNeighbors(n_neighbors=1, algorithm='ball_tree', metric=metric).fit(linear_codes)
        metric_output, indices = nn.kneighbors(linear_features)
        gene_ids = np.ravel(norm_codes.indexes[Features.TARGET].values[indices])

        return np.ravel(metric_output), gene_ids

    def _validate_decode_intensity_input_matches_codebook_shape(
            self,
            intensities: IntensityTable,
    ):
        # verify that the shapes of the codebook and intensities match
        ch_match = intensities.sizes[Indices.CH] == self.sizes[Indices.CH]
        round_match = intensities.sizes[Indices.ROUND] == self.sizes[Indices.ROUND]
        if not (ch_match and round_match):
            raise ValueError(
                'Codebook and Intensities must have same number of channels and rounds')

    def metric_decode(
            self, intensities: IntensityTable, max_distance: Number, min_intensity: Number,
            norm_order: int, metric: str='euclidean'
    ) -> IntensityTable:
        """Assign the closest target by euclidean distance to each feature in an intensity table

        Normalizes both the codes and the features to be unit vectors and finds the closest code
        for each feature

        Parameters
        ----------
        intensities : IntensityTable
            features to be decoded
        max_distance : Number
            maximum distance between a feature and its closest code for which the coded target will
            be assigned.
        min_intensity : Number
            minimum intensity for a feature to receive a target annotation
        norm_order : int
            the scipy.linalg norm to apply to normalize codes and intensities
        metric : str
            the sklearn metric string to pass to NearestNeighbors

        See Also
        --------
        The available norms for this function can be found at the following link:
        https://docs.scipy.org/doc/numpy-1.14.0/reference/generated/numpy.linalg.norm.html

        Returns
        -------
        IntensityTable :
            Intensity table containing normalized intensities, target assignments, distances to
            the nearest code, and the filtering status of each feature.

        """

        self._validate_decode_intensity_input_matches_codebook_shape(intensities)

        # normalize both the intensities and the codebook
        norm_intensities, norms = self._normalize_features(intensities, norm_order=norm_order)
        norm_codes, _ = self._normalize_features(self, norm_order=norm_order)

        metric_outputs, targets = self._approximate_nearest_code(
            norm_codes, norm_intensities, metric=metric)

        # only targets with low distances and high intensities should be retained
        passes_filters = np.logical_and(
            norms >= min_intensity,
            metric_outputs <= max_distance,
            dtype=np.bool
        )

        # set targets, distances, and filtering results
        norm_intensities[Features.TARGET] = (Features.AXIS, targets)
        norm_intensities[Features.DISTANCE] = (Features.AXIS, metric_outputs)
        norm_intensities[Features.PASSES_THRESHOLDS] = (Features.AXIS, passes_filters)

        # norm_intensities is a DataArray, make it back into an IntensityTable
        return IntensityTable(norm_intensities)

    def decode_per_round_max(self, intensities: IntensityTable) -> IntensityTable:
        """decode each feature by selecting the per-imaging-round max-valued channel

        Notes
        -----
        - If no code matches the per-round maximum for a feature, it will be assigned 'nan' instead
          of a target value
        - Numpy's argmax breaks ties by picking the first channel -- this can lead to
          unexpected results where some features with "tied" channels will decode, but others will
          be assigned 'nan'.

        Parameters
        ----------
        intensities : IntensityTable
            features to be decoded

        Returns
        -------
        IntensityTable :
            intensity table containing additional data variables for target assignments

        """

        def _view_row_as_element(array: np.ndarray) -> np.ndarray:
            """view an entire code as a single element

            This view allows vectors (codes) to be compared for equality without need for multiple
            comparisons by casting the data in each code to a structured dtype that registers as
            a single value

            Parameters
            ----------
            array : np.ndarray
                2-dimensional numpy array of shape (n_observations, (n_ch * n_round)) where
                observations may be either features or codes.

            Returns
            -------
            np.ndarray :
                1-dimensional vector of shape n_observations

            """
            nrows, ncols = array.shape
            dtype = {'names': ['f{}'.format(i) for i in range(ncols)],
                     'formats': ncols * [array.dtype]}
            return array.view(dtype)

        self._validate_decode_intensity_input_matches_codebook_shape(intensities)

        max_channels = intensities.argmax(Indices.CH.value)
        codes = self.argmax(Indices.CH.value)

        # TODO ambrosejcarr, dganguli: explore this quality score further
        # calculate distance scores by evaluating the fraction of signal in each round that is
        # found in the non-maximal channels.
        max_intensities = intensities.max(Indices.CH.value)
        round_intensities = intensities.sum(Indices.CH.value)
        distance = 1 - (max_intensities / round_intensities).mean(Indices.ROUND.value)

        a = _view_row_as_element(codes.values.reshape(self.shape[0], -1))
        b = _view_row_as_element(max_channels.values.reshape(intensities.shape[0], -1))

        targets = np.full(intensities.shape[0], fill_value=np.nan, dtype=object)

        # decode the intensities
        for i in np.arange(codes.shape[0]):
            targets[np.where(a[i] == b)[0]] = codes[Features.TARGET][i]

        # a code passes filters if it decodes successfully
        passes_filters = ~pd.isnull(targets)

        intensities[Features.TARGET] = (Features.AXIS, targets.astype('U'))
        intensities[Features.DISTANCE] = (Features.AXIS, distance)
        intensities[Features.PASSES_THRESHOLDS] = (Features.AXIS, passes_filters)

        return intensities

    @classmethod
    def synthetic_one_hot_codebook(
            cls, n_round: int, n_channel: int, n_codes: int, target_names: Optional[Sequence]=None
    ) -> "Codebook":
        """Generate codes where one channel is "on" in each imaging round

        Parameters
        ----------
        n_round : int
            number of imaging rounds per code
        n_channel : int
            number of channels per code
        n_codes : int
            number of codes to generate
        target_names : Optional[List[str]]
            if provided, names for targets in codebook

        Examples
        --------
        Create a Codebook with 2 rounds, 3 channels, and 2 codes::

            >>> from starfish import Codebook
            >>> Codebook.synthetic_one_hot_codebook(n_round=2, n_channel=3, n_codes=2)
            <xarray.Codebook (target: 2, c: 3, h: 2)>
            array([[[0, 1],
                    [0, 0],
                    [1, 0]],

                   [[1, 1],
                    [0, 0],
                    [0, 0]]], dtype=uint8)
            Coordinates:
              * target     (target) object b25180dc-8af5-48f1-bff4-b5649683516d ...
              * c          (c) int64 0 1 2
              * h          (h) int64 0 1

        Returns
        -------
        List[Dict] :
            list of codewords

        """

        # TODO ambrosejcarr: clean up this code, generate Codebooks directly using _empty_codebook
        # construct codes; this can be slow when n_codes is large and n_codes ~= n_possible_codes
        codes: Set = set()
        while len(codes) < n_codes:
            codes.add(tuple([np.random.randint(0, n_channel) for _ in np.arange(n_round)]))

        # construct codewords from code
        codewords = [
            [
                {
                    Indices.ROUND.value: h, Indices.CH.value: c, 'v': 1
                } for h, c in enumerate(code)
            ] for code in codes
        ]

        # make a codebook from codewords
        if target_names is None:
            # use a reverse-sorted list of integers as codewords
            target_names = [uuid.uuid4() for _ in range(n_codes)]
        assert n_codes == len(target_names)

        codebook = [{Features.CODEWORD: w, Features.TARGET: g}
                    for w, g in zip(codewords, target_names)]

        return cls.from_code_array(codebook, n_round=n_round, n_ch=n_channel)



import click

from sptx_format import validate_sptx


@click.command()
@click.option("--experiment-json", required=True, metavar="JSON_FILE_OR_URL")
@click.option("--fuzz", is_flag=True)
@click.pass_context
def validate(ctx, experiment_json, fuzz):
    """invokes validate with the parsed commandline arguments"""
    try:
        valid = validate_sptx.validate(experiment_json, fuzz)
        if valid:
            ctx.exit(0)
        else:
            ctx.exit(1)
    except KeyboardInterrupt:
        ctx.exit(3)



"""
Tests for IntensityTable.synthetic_intensities method.
"""

import numpy as np

from starfish import IntensityTable
from starfish.test import test_utils
from starfish.types import Features, Indices


def test_synthetic_intensity_generation():
    """
    Create a 2-spot IntensityTable of pixel size (z=3, y=4, x=5) from a codebook with 3 channels
    and 2 rounds.

    Verify that the constructed Synthetic IntensityTable conforms to those dimensions, and given
    a known random seed, that the output spots decode to match a target in the input Codebook
    """
    # set seed to check that codebook is matched. This seed generates 2 instances of GENE_B
    np.random.seed(1)
    codebook = test_utils.codebook_array_factory()
    num_z, height, width = 3, 4, 5
    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=num_z,
        height=height,
        width=width,
        n_spots=2
    )

    # sizes should match codebook
    assert intensities.sizes[Indices.ROUND] == 2
    assert intensities.sizes[Indices.CH] == 3
    assert intensities.sizes[Features.AXIS] == 2

    # attributes should be bounded by the specified size
    assert np.all(intensities[Indices.Z.value] <= num_z)
    assert np.all(intensities[Indices.Y.value] <= height)
    assert np.all(intensities[Indices.X.value] <= width)

    # both codes should match GENE_B
    assert np.array_equal(
        np.where(intensities.values),
        [[0, 0, 1, 1],  # two each in feature 0 & 1
         [1, 2, 1, 2],  # one each in channel 1 & 2
         [1, 0, 1, 0]],  # channel 1 matches round 1, channel 2 matches round zero
    )



"""
This module describes the contracts to provide data to the experiment builder.
"""

from typing import Mapping, Tuple, Union

import numpy as np
from slicedimage import (
    ImageFormat,
)

from starfish.types import Coordinates, Number


class FetchedTile:
    """
    This is the contract for providing the data for constructing a :class:`slicedimage.Tile`.
    """
    def __init__(self, *args, **kwargs):
        pass

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return Tile shape.

        Returns
        -------
        Tuple[int, ...]
            The tile shape in (y, x)
        """
        raise NotImplementedError()

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        """Return the tile's coordinates in the global coordinate space..

        Returns
        -------
        Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]
            Maps from a coordinate type (e.g. 'x', 'y', or 'z') to its value or range.
        """
        raise NotImplementedError()

    @property
    def extras(self) -> dict:
        """Return the extras data associated with the tile.

        Returns
        -------
        Mapping[str, Any]
            Maps from a key to its value.
        """
        return {}

    @property
    def format(self) -> ImageFormat:
        """Return the Tile's format

        Returns
        -------
        ImageFormat :
            a slicedimage format type, e.g. ImageFormat.TIFF
        """
        raise NotImplementedError()

    def tile_data(self) -> np.ndarray:
        """Return the image data representing the tile.

        Returns
        -------
        ndarray :
            The image data
        """
        raise NotImplementedError()


class TileFetcher:
    """
    This is the contract for providing the image data for constructing a
    :class:`slicedimage.Collection`.
    """
    def get_tile(self, fov: int, r: int, ch: int, z: int) -> FetchedTile:
        """
        Given fov, r, ch, and z, return an instance of a :class:`.FetchedImage` that can be
        queried for the image data.
        """
        raise NotImplementedError()
